# Copyright 2017 James P Goodwin chardraw unicode curses based graphics package
""" module that implements a graphics package using block graphics on curses window """
import locale
locale.setlocale(locale.LC_ALL,"")
import curses
import curses.ascii
import sys
import os
import math

class Canvas:
    """ primitive drawing surface attached to a curses window """

    to_mask = { 0:1, 1:2, 2:4, 3:8 }
    mask_to_char = {
        0 :'\u2008',
        1 :'\u2598',
        2 :'\u259d',
        3 :'\u2580',
        4 :'\u2596',
        5 :'\u258c',
        6 :'\u259e',
        7 :'\u259b',
        8 :'\u2597',
        9 :'\u259a',
        10:'\u2590',
        11:'\u259c',
        12:'\u2584',
        13:'\u2599',
        14:'\u259f',
        15:'\u2588'
        }

    char_to_mask = {
        '\u2008':0 ,
        '\u2598':1 ,
        '\u259d':2 ,
        '\u2580':3 ,
        '\u2596':4 ,
        '\u258c':5 ,
        '\u259e':6 ,
        '\u259b':7 ,
        '\u2597':8 ,
        '\u259a':9 ,
        '\u2590':10,
        '\u259c':11,
        '\u2584':12,
        '\u2599':13,
        '\u259f':14,
        '\u2588':15
        }

    def __init__(self, win = None ):
        """ constructor, can be initialized with a window to draw on, otherwise window must be set later by set_window """
        self.set_win(win)

    def to_rowcol(self, x, y ):
        """ return character row col for input x,y coordinates """
        return (int(y/2),int(x/2))

    def from_rowcol(self, row, col ):
        """ return the pixel location of a character position, returns upper left pixel in matrix"""
        return (int(row)*2,int(col)*2)

    def round_text_position(self, x, y):
        """ adjust a text position so that it always ends up down and to the right if it is at a half pixel offset """
        r,c = self.to_rowcol(x,y)
        y1,x1 = self.from_rowcol(r,c)
        h,w = self.from_rowcol(1,1)
        if y1 < y:
            y = y + h/2
        if x1 < x:
            x = x + w/2
        return x, y

    def round_text_x_position(self, x):
        """ adjust a text position so that it always ends up down and to the right if it is at a half pixel offset """
        r,c = self.to_rowcol(x,0)
        y1,x1 = self.from_rowcol(r,c)
        h,w = self.from_rowcol(1,1)
        if x1 < x:
            x = x + w/2
        return x

    def round_text_y_position(self, y):
        """ adjust a text position so that it always ends up down and to the right if it is at a half pixel offset """
        r,c = self.to_rowcol(0,y)
        y1,x1 = self.from_rowcol(r,c)
        h,w = self.from_rowcol(1,1)
        if y1 < y:
            y = y + h/2
        return y

    def set_win(self, win ):
        """ point this canvas at a window and initialize things, will blank out the window """
        self.win = win
        self.init_win()

    def init_win(self):
        """ initializes the window and sets up all of the defaults """
        curses.init_pair(5,curses.COLOR_BLACK,curses.COLOR_BLACK)
        curses.init_pair(1,curses.COLOR_GREEN,curses.COLOR_BLACK)
        curses.init_pair(2,curses.COLOR_RED,curses.COLOR_BLACK)
        curses.init_pair(3,curses.COLOR_CYAN,curses.COLOR_BLACK)
        curses.init_pair(4,curses.COLOR_WHITE,curses.COLOR_BLACK)

        self.green = curses.color_pair(1)
        self.red = curses.color_pair(2)
        self.cyan = curses.color_pair(3)
        self.white = curses.color_pair(4)
        self.black = curses.color_pair(5)

        if curses.can_change_color():
            self.color_min = 8
            self.color_max = 256

            red = 0
            green = 100
            blue = 20
            for c in range(self.color_min,self.color_max):
                curses.init_color(c,red,green,blue)
                red += 23
                green += 33
                blue += 53
                red = red % 1000
                green = green % 1000
                blue = blue % 1000

            for cidx in range(self.color_min,self.color_max):
                curses.init_pair(cidx,cidx,curses.COLOR_BLACK)
        else:
            self.color_min = 0
            self.color_max = 8

        if self.win:
            self.max_y,self.max_x = self.win.getmaxyx()
            self.char_map = [[None] * self.max_y for i in range(self.max_x)]
            self.max_y = self.max_y * 2
            self.max_x = self.max_x * 2
        else:
            self.max_y,self.max_x = (0,0)
            self.char_map = None

    def refresh( self ):
        """ refresh the display after drawing """
        self.win.refresh()

    def get_maxxy( self ):
        """ return the maximum number of x and y pixels that are available in this canvas """
        return (self.max_x,self.max_y)

    def put_pixel( self, x,y, color, set = True ):
        """ turn on a pixel with the color indicated """
        if x < 0 or x >= self.max_x or y < 0 or y >= self.max_y:
            return
        row,col = self.to_rowcol(x,y)
        mask = self.to_mask[(int(x)%2)+((int(y)%2)*2)]

        if not self.char_map[col][row]:
            current_mask = 0
        else:
            current_mask = self.char_to_mask[self.char_map[col][row]]

        if set:
            self.char_map[col][row] = self.mask_to_char[ mask | current_mask ]
        else:
            self.char_map[col][row] = self.mask_to_char[ mask ^ current_mask ]
        try:
            self.win.addstr(row,col,self.char_map[col][row].encode('utf_8'),color)
        except:
            pass

    def line(self, x0, y0, x1, y1, color, put_pixel=None ):
        """ draw a line between x0,y0 and x1,y1 in color """
        x0 = int(x0)
        x1 = int(x1)
        y0 = int(y0)
        y1 = int(y1)

        dx = abs(x1 - x0)
        dy = abs(y1 - y0)
        x, y = x0, y0
        sx = -1 if x0 > x1 else 1
        sy = -1 if y0 > y1 else 1
        if dx > dy:
            err = dx / 2.0
            while x != x1:
                if put_pixel:
                    put_pixel(x,y,color)
                else:
                    self.put_pixel(x, y, color)
                err -= dy
                if err < 0:
                    y += sy
                    err += dx
                x += sx
        else:
            err = dy / 2.0
            while y != y1:
                if put_pixel:
                    put_pixel(x,y,color)
                else:
                    self.put_pixel(x, y, color)
                err -= dx
                if err < 0:
                    x += sx
                    err += dy
                y += sy
        if put_pixel:
            put_pixel(x, y, color)
        else:
            self.put_pixel(x,y,color)

    def intersect( self, seg1, seg2, clip_to_seg = False ):
        """ find the intersection of two segments as tuples (x0,y0,x1,y1) returns tuple (x,y) if no intersection returns None """
        def lineform( seg ):
            """ return A, B, C for the standard line formula Ax + By = C """
            A = float(seg[1]-seg[3])
            B = float(seg[2]-seg[0])
            C = A*seg[0]+B*seg[1]
            return (A,B,C)

        l1 = lineform(seg1)
        l2 = lineform(seg2)

        det = l1[0]*l2[1] - l2[0]*l1[1]
        if det != 0:
            x = (l2[1]*l1[2] - l1[1]*l2[2])/det
            y = (l1[0]*l2[2] - l2[0]*l1[2])/det

            if clip_to_seg:
                if x >= min(seg1[0],seg1[2]) and x <= max(seg1[0],seg1[2]) and y >= min(seg1[1],seg1[3]) and y <= max(seg1[1],seg1[3]):
                    return (int(x),int(y))
            else:
                return (int(x),int(y))
        return None

    def cross_product_length( self, pA,pB,pC ):
        """ compute the cross product of AB x BC """
        BAx = float(pA[0] - pB[0])
        BAy = float(pA[1] - pB[1])
        BCx = float(pC[0] - pB[0])
        BCy = float(pC[1] - pB[1])

        return (BAx * BCy - BAy * BCx)

    def is_convex( self, points ):
        """ take a list of (x,y) tuples representing the vertecies of a polygon in order and return True if it represents a convex polygon, False otherwise """
        got_negative = False
        got_positive = False
        num_points = len(points)
        if num_points <= 3:
            return True


        min_x,min_y,max_x,max_y = self.get_bounds(points)
        if max_x-min_x <= 1.0 or max_y-min_y <= 1.0:
            return True

        for A in range(num_points):
            B = (A+1)%num_points
            C = (B+1)%num_points
            cross_product = self.cross_product_length(points[A],points[B],points[C])
            if cross_product < 0:
                got_negative = True
            elif cross_product > 0:
                got_positive = True

        return not (got_negative and got_positive)

    def get_bounds(self,points):
        """ return tuple (min_x,min_y,max_x,max_y) for list of points """
        min_x = -1
        min_y = -1
        max_x = -1
        max_y = -1
        for x,y in points:
            if min_x < 0 or x < min_x:
                min_x = x
            if min_y < 0 or y < min_y:
                min_y = y
            if max_x < 0 or x > max_x:
                max_x = x
            if max_y < 0 or y > max_y:
                max_y = y
        return (min_x,min_y,max_x,max_y)

    def clip_polygon(self, points, minX, minY, maxX, maxY, dir=-1 ):
        """ clip a polygon against the bounds exressed by minX,minY to maxX,maxY and return either None for nothing inside or the points for the polygon dir is -1 all,0=top,1=right,2=bottom,3=left """
        def inside( p, minX, minY, maxX, maxY, dir ):
            x,y = p
            if dir == 0:
                return(y >= minY)
            elif dir == 1:
                return(x < maxX)
            elif dir == 2:
                return(y < maxY)
            elif dir == 3:
                return(x >= minX)

        def intersect(sp, ep, minX, minY, maxX, maxY, dir ):
            x0,y0 = sp
            x1,y1 = ep
            s1 = (x0,y0,x1,y1)
            if dir == 0:
                s2 = (minX,minY,maxX,minY)
            elif dir == 1:
                s2 = (maxX,minY,maxX,maxY)
            elif dir == 2:
                s2 = (minX,maxY,maxX,maxY)
            elif dir == 3:
                s2 = (minX,minY,minX,maxY)
            return self.intersect(s1,s2,False)

        if dir == -1:
            for d in [0,1,2,3]:
                points = self.clip_polygon(points,minX,minY,maxX,maxY,d)
                if not points:
                    return None
            return points
        else:
            sp = points[-1]
            out_points = []
            for ep in points:
                if inside(ep,minX,minY,maxX,maxY,dir):
                    if inside(sp,minX,minY,maxX,maxY,dir):
                        out_points.append(ep)
                    else:
                        ip = intersect(sp,ep,minX,minY,maxX,maxY,dir)
                        out_points.append(ip)
                        out_points.append(ep)
                else:
                    if inside(sp,minX,minY,maxX,maxY,dir):
                        ip = intersect(sp,ep,minX,minY,maxX,maxY,dir)
                        out_points.append(ip)
                sp = ep
            return out_points if out_points else None


    def rasterize( self, points, color, put_pixel=None):
        """ sort points representing the boundary of a filled shape and rasterize by filling lines with color """
        ps = sorted(points,key=lambda x: (x[1],x[0]))
        sp = iter(ps)
        p2 = p1 = next(sp,None)
        si = 0
        last = None
        while p2:
            last = p2
            p2 = next(sp,None)
            if not p2:
                if put_pixel:
                    put_pixel(p1[0],p1[1],color)
                else:
                    self.put_pixel(p1[0],p1[1],color)
            else:
                if p2[1] == p1[1]:
                    continue
                else:
                    self.line(p1[0],p1[1],last[0],last[1],color,put_pixel)
                    p1 = p2



    def circle(self, x0, y0, radius, color, fill = False, put_pixel=None ):
        """ draw a circle centered at x0,y0 of radius radius in color """
        self.arc(x0,y0,radius,0,360,color,fill,put_pixel)

    def arc(self,x0,y0,radius,a0,a1,color,fill=False,put_pixel=None,just_points=False):
        """ draw an arc between a0 degrees to a1 degrees centered at x0,y0 with radius and color """
        def circle_point(x0,y0,a,radius):
            return (x0+(1.5*math.cos(math.radians(a))*radius), y0+math.sin(math.radians(a))*radius)

        points = []
        points.append((x0,y0))
        a = a0
        while a <= a1:
            xp,yp = circle_point(x0,y0,a,radius)
            a = a + 1.0
            points.append((xp,yp))
                                

        if just_points:
            for x,y in points:
                put_pixel(x,y,color)
        else:
            self.polygon(points,color,fill,put_pixel)

    def rect(self,x0,y0,x1,y1,color,fill=False,put_pixel=None):
        """ draw a rectangle bounding x0,y0, x1,y1, in color == color optionally filling """
        x0 = int(x0)
        x1 = int(x1)
        y0 = int(y0)
        y1 = int(y1)

        if not fill:
            self.line(x0,y0,x0,y1,color)
            self.line(x0,y1,x1,y1,color)
            self.line(x1,y1,x1,y0,color)
            self.line(x1,y0,x0,y0,color)
        else:
            if y1 < y0:
                y=y0
                y0=y1
                y1 = y
            for y in range(y0,y1):
                self.line(x0,y,x1,y,color,put_pixel)

    def textat(self,x,y,color,message):
        """ draw a text message at a coordinate in the color specified """
        x,y = self.round_text_position(x,y)

        height, width = self.from_rowcol(1,len(message))

        if x < 0 or x >self.max_x or y < 0 or y >self.max_y:
            return

        if y + height > self.max_y:
            return

        if x + height > self.max_x:
            clip_height,clip_width = self.to_rowcol(1,(self.max_x-x))
            if clip_width > 0:
                message = message[:clip_width]
            else:
                return

        row,col = self.to_rowcol(x,y)
        self.win.addstr(row,col,message.encode('utf_8'),color)

    def polyline(self,points,color,put_pixel=None):
        """ draw a polyline defined by the sequence points which represent a list of (x,y) tuples in the order they should be connected in color """
        i = iter(points)
        p1 = next(i,None)
        while p1:
            p2 = next(i,None)
            if p2:
                self.line(p1[0],p1[1],p2[0],p2[1],color,put_pixel)
            else:
                if put_pixel:
                    put_pixel(p1[0],p1[1],color)
                else:
                    self.put_pixel(p1[0],p1[1],color)
            p1 = p2

    def poly_fill(self,points,color,put_pixel = None):
        """ fill a concave polygon by recursively subdividing until we get a convex polygon """
        clips = []

        minX,minY,maxX,maxY = self.get_bounds(points)

        minX = float(minX)
        minY = float(minY)
        maxX = float(maxX)
        maxY = float(maxY)

        midX = (minX+maxX)/2.0
        midY = (minY+maxY)/2.0

        clips.append((minX,minY,midX,midY))
        clips.append((midX,minY,maxX,midY))
        clips.append((midX,midY,maxX,maxY))
        clips.append((minX,midY,midX,maxY))

        while clips:
            minX,minY,maxX,maxY = clips.pop(0)
            if int(minX)==int(maxX) or int(minY)==int(maxY):
                continue

            p = self.clip_polygon(points,minX,minY,maxX,maxY)
            if p:
                if self.is_convex(p):
                    self.polygon(p,color,True,put_pixel)
                else:
                    midX = (minX+maxX)/2.0
                    midY = (minY+maxY)/2.0
                    if midX - minX < 1.0 or midY - minY < 1.0 or maxX - midX < 1.0 or maxY - midY < 1.0:
                        continue
                    clips.append((minX,minY,midX,midY))
                    clips.append((midX,minY,maxX,midY))
                    clips.append((midX,midY,maxX,maxY))
                    clips.append((minX,midY,midX,maxY))

    def polygon(self,points,color,fill=False,put_pixel=None):
        """ draw a polygon defined by the sequence points which represent a list of (x,y) tuples in the order they should be connected in color
        the last point will be connected to the first point. polygons can be filled. """

        if not points:
            return

        convex = True
        if fill:
            convex = self.is_convex(points)

        poly_pixels = []
        def put_poly_pixel(x,y,color):
            poly_pixels.append((x,y))

        i = iter(points)
        first = p1 = next(i,None)
        while p1:
            p2 = next(i,None)
            if p2:
                last = p2
                self.line(p1[0],p1[1],p2[0],p2[1],color,put_poly_pixel)
            else:
                last = p1
                put_poly_pixel(p1[0],p1[1],color)
            p1 = p2
        self.line(first[0],first[1],last[0],last[1],color,put_poly_pixel)

        if not fill:
            for x,y in poly_pixels:
                if put_pixel:
                    put_pixel(x,y,color)
                else:
                    self.put_pixel(x,y,color)
        else:
            if convex:
                self.rasterize( poly_pixels, color, put_pixel)
            else:
                for x,y in poly_pixels:
                    if put_pixel:
                        put_pixel(x,y,color)
                    else:
                        self.put_pixel(x,y,color)
                self.poly_fill(points,color,put_pixel)



# Copyright 2020 James P Goodwin data table package to manage sparse columnar data
""" module that aggregates data from the syslog and provides a set of data tables """
import locale
locale.setlocale(locale.LC_ALL,'')
import sys
import os
import glob
import gzip
import re
from datetime import datetime,timedelta
from data_sources.data_table import DataTable,Column,Cell,blank_type,string_type,float_type,int_type,date_type,format_string,format_float,format_date,format_int,synchronized

class SyslogDataTable( DataTable ):
    """ class that collects a time based aggregation of data from the syslog into a data_table """
    def __init__(self,syslog_glob="/var/log/syslog*",num_hours=24,bucket_hours=1,refresh_minutes=10):
        """ Initialize the SyslogDataTable with a file glob pattern to collect the syslogs on this machine, a timespan to aggregate for, aggregation bucket in hours, a refresh interval for updating in minutes """
        self.syslog_glob = syslog_glob
        self.num_hours = num_hours
        self.bucket_hours = bucket_hours
        DataTable.__init__(self,None,
            "Syslog Data: %s for the last %d hours in %d hour buckets, refreshed every %d minutes"%(
            self.syslog_glob,
            self.num_hours,
            self.bucket_hours,
            refresh_minutes),refresh_minutes)
        self.refresh()

    @synchronized
    def refresh( self ):
        """ refresh or rebuild tables """
        current_time = datetime.now()
        start_time = current_time - timedelta( hours = self.num_hours )
        syslog_files = glob.glob(self.syslog_glob)

        time_column = Column(name="Time Stamps")
        bucket_time = start_time
        idx = 0
        while bucket_time < current_time:
            time_column.put(idx,Cell(date_type,bucket_time,format_date))
            bucket_time = bucket_time + timedelta( hours = self.bucket_hours )
            idx += 1
        time_column.put(idx,Cell(date_type,current_time,format_date))

        def bucket_idx( timestamp ):
            if timestamp < start_time or timestamp > current_time:
                return -1

            for idx in range(time_column.size()):
                if time_column.get(idx).get_value() >= timestamp:
                    return idx
            else:
                return -1

        errors_column = Column(name="Errors by Time")
        warnings_column = Column(name="Warnings by Time")
        messages_column = Column(name="Messages by Time")

        services_column = Column(name="Services")
        errors_service_column = Column(name="Errors by Service")
        warnings_service_column = Column(name="Warnings by Service")
        messages_service_column = Column(name="Messages by Service")

        def service_idx( service ):
            for idx in range(services_column.size()):
                if services_column.get(idx).get_value() == service:
                    return idx
            else:
                return -1

        def put_or_sum( column, idx, value ):
            current_value = 0
            if idx < column.size():
                c = column.get(idx)
                if c.get_type() != blank_type:
                    current_value = int(c.get_value())
            column.put(idx,Cell(int_type,current_value+value,format_int))


        for slf in syslog_files:
            if slf.endswith(".gz"):
                slf_f = gzip.open(slf,"rt",encoding="utf-8")
            else:
                slf_f = open(slf,"r",encoding="utf-8")

            for line in slf_f:
                line = line.strip()
                m = re.match(r"(\w\w\w\s+\d+\s\d\d:\d\d:\d\d)\s[a-z0-9\-]*\s([a-zA-Z0-9\-\_\.]*)[\[\]0-9]*:\s*(.*)",line)
                if m:
                    log_date = re.sub("\s+"," ","%d "%current_time.year + m.group(1))
                    log_process = m.group(2)
                    log_message = m.group(3)
                    log_datetime = datetime.strptime(log_date,"%Y %b %d %H:%M:%S")
                    b_idx = bucket_idx( log_datetime )
                    if b_idx >= 0:
                        s_idx = service_idx( log_process )
                        if s_idx < 0:
                            s_idx = services_column.size()
                            services_column.put(s_idx,Cell(string_type,log_process,format_string))
                        put_or_sum(messages_column,b_idx,1)
                        put_or_sum(messages_service_column,s_idx,1)
                        is_error = re.search("[Ee]rror|ERROR",log_message)
                        is_warning = re.search("[Ww]arning|WARNING",log_message)
                        error_count = 0
                        warning_count = 0
                        if is_error and not is_warning:
                            error_count = 1
                        elif is_warning:
                            warning_count = 1
                        put_or_sum(errors_column,b_idx,error_count)
                        put_or_sum(errors_service_column,s_idx,error_count)
                        put_or_sum(warnings_column,b_idx,warning_count)
                        put_or_sum(warnings_service_column,s_idx,warning_count)

        columns = [time_column,errors_column,warnings_column,messages_column,services_column,
                    errors_service_column,warnings_service_column,messages_service_column]

        for c in columns:
            if self.has_column(c.get_name()):
                self.replace_column(self.map_column(c.get_name()),c)
            else:
                self.add_column(c)

        self.changed()

        DataTable.refresh(self)



# Copyright 2020 James P Goodwin data table package to manage sparse columnar data
""" module that aggregates data from the syslog and provides a set of data tables """
import locale
locale.setlocale(locale.LC_ALL,'')
import sys
import os
import glob
import gzip
import re
from elasticsearch import Elasticsearch
from datetime import datetime,timedelta
from data_sources.data_table import DataTable,Column,Cell,blank_type,string_type,float_type,int_type,date_type,format_string,format_float,format_date,format_int,synchronized

class ElasticsearchDataTable( DataTable ):
    """ class that collects data from the response to a specific elasticsearch query and populates tables based on a field map """
    def __init__(self,refresh_minutes=1,es_index_pattern=None,es_query_body=None,es_field_map=None):
        """ Initialize the ElasticsearchQueryTable pass in a refresh interval in minutes, the es_query_body dict representing the query json and the field map list of tuples [( json path, field name, field type )...]"""
        self.es_query_body = es_query_body
        self.es_field_map = es_field_map
        self.es_index_pattern = es_index_pattern
        DataTable.__init__(self,None,
            "Elasticsearch query:%s,index:%s,fieldmap:%s,refreshed every %d minutes"%(
                es_query_body,es_index_pattern,es_field_map,refresh_minutes),
            refresh_minutes)
        self.refresh()

    @synchronized
    def get_es_parameters( self ):
        """ fetch the elasticsearch parameters for this table as a tuple (es_query_body,es_field_map,es_index_pattern) """
        return (self.es_query_body,self.es_field_map,self.es_index_pattern)

    @synchronized
    def set_es_parameters( self, es_query_body,es_field_map,es_index_pattern ):
        """ set the elasticsearch parameters for this table """
        self.es_query_body = es_query_body
        self.es_field_map = es_field_map
        self.es_index_pattern = es_index_pattern
        self.refresh()

    @synchronized
    def refresh( self ):
        """ refresh or rebuild tables """

        es = Elasticsearch()

        result = es.search(index=self.es_index_pattern,body=self.es_query_body)

        def match_fields( name, result ):
            matches = []
            if isinstance(result,dict):
                for k in result:
                    full_name = (name+"." if name else "")+k
                    item = result[k]
                    for json_path,field_name,field_type  in self.es_field_map:
                        if full_name == json_path:
                            matches.append((field_name,field_type,item))
                    if isinstance(item,dict) or isinstance(item,list):
                        matches += match_fields(full_name,item)
            elif isinstance(result,list):
                for k in result:
                    matches += match_fields(name,k)
            return matches

        matches = match_fields( "",result)

        new_columns = {}
        for column_name,column_type,value in matches:
            if not column_name in new_columns:
                new_columns[column_name] = Column(name=column_name)
            c = new_columns[column_name]
            if column_type == "date":
                cc = Cell(date_type,datetime.fromtimestamp(value/1000),format_date)
            elif column_type == "int":
                cc = Cell(int_type,value,format_int)
            elif column_type == "float":
                cc = Cell(float_type,value,format_float)
            elif column_type == "str":
                cc = Cell(string_type,value,format_string)
            else:
                cc = Cell(string_type,str(value),format_string)
            c.put(c.size(),cc)

        for column_name in new_columns:
            if self.has_column(column_name):
                self.replace_column(self.map_column(column_name),new_columns[column_name])
            else:
                self.add_column(new_columns[column_name])

        self.changed()

        DataTable.refresh(self)



# Copyright 2017 James P Goodwin data table package to manage sparse columnar data
""" module that implement a data table package to manage sparse columnar data window """
import sys
import os
import re
from datetime import datetime
import threading
import time
import csv
import json
from io import StringIO
from paramiko.client import SSHClient
import keyring
from functools import wraps
from data_sources.data_table import DataTable,Cell,Column,from_json,to_json,synchronized
from dashboard.version import __version__

def sync_connection(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        with self.connection_lock:
            return method(self, *args, **kwargs)
    return wrapper

def sync_manager(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        with self.manager_lock:
            return method(self, *args, **kwargs)
    return wrapper

class Connection():
    def __init__(self, owner=None, ssh_client=None, session=None, stdin=None, stdout=None, stderr=None ):
        """ provides protocol to the table server """
        self.ssh_client = ssh_client
        self.session = session
        self.clients = []
        self.stdout_lines = []
        self.stderr_lines = []
        self.owner = owner
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.connection_lock = threading.RLock()
        self.reader_lock = threading.RLock()
        self.stdout_reader_thread = threading.Thread(target=self.reader,args=(self.stdout,self.stdout_lines))
        self.stdout_reader_thread.start()
        self.stderr_reader_thread = threading.Thread(target=self.reader,args=(self.stderr,self.stderr_lines))
        self.stderr_reader_thread.start()

    def reader( self, stream, lines ):
        """ worker thread that reads from stdout and pushes data onto stdout_lines and stderr_lines """
        while not self.session.exit_status_ready():
            line = stream.readline()
            with self.reader_lock:
                lines.append(line)

        with self.reader_lock:
            lines += stream.readlines()

    @sync_connection
    def get_stdout_line( self ):
        """ fetch a line from the queue of stdout_lines """
        while True:
            with self.reader_lock:
                if len(self.stdout_lines):
                    return self.stdout_lines.pop(0)
            time.sleep(0)

    @sync_connection
    def get_stderr_line( self ):
        """ fetch a line from the queue of stderr_lines """
        while True:
            with self.reader_lock:
                if len(self.stderr_lines):
                    return self.stderr_lines.pop(0)
            time.sleep(0)

    @sync_connection
    def open( self, client ):
        """ register this client as a user of this connection """
        if client not in self.clients:
            self.clients.append(client)

    @sync_connection
    def table(self, table_def):
        """ send request to create a new remote table, returns loaded response """
        print("table:%s"%table_def,file=self.stdin,flush=True)
        return self.get_stdout_line()

    @sync_connection
    def refresh(self, table_name):
        """ send request to refresh a remote table named table_name and return response """
        print("refresh:%s"%table_name,file=self.stdin,flush=True)
        return self.get_stdout_line()

    @sync_connection
    def get(self, table_name):
        """ send request to fetch table_name and return response """
        print("get:%s"%table_name,file=self.stdin,flush=True)
        return self.get_stdout_line()

    @sync_connection
    def exit(self):
        """ terminate the server and clean up this connection """
        print("exit",file=self.stdin,flush=True)
        return ""

    @sync_connection
    def close(self,client):
        """ close this client's use of this connection """
        if client in self.clients:
            self.clients.remove(client)

class ConnectionManager():
    def __init__(self):
        """ manages connections to servers and their initial setup """
        self.connections = {}
        self.manager_lock = threading.RLock()

    def __del__(self):
        """ just in case clean up any connections """
        self.shutdown()

    @sync_manager
    def shutdown( self ):
        """ shut down the connection manager and close all the pooled connections """
        for cn in self.connections:
            self.connections[cn].exit()
            self.connections[cn].ssh_client.close()
        self.connections = {}

    @sync_manager
    def connect(self,ssh_spec,client):
        """ create a connection to a server """
        if ssh_spec in self.connections:
            connection = self.connections[ssh_spec]
            connection.open(client)
            return connection

        username,server,port = re.match(r"ssh://([a-z_][a-z0-9_-]*\${0,1})@([^:]*):{0,1}(\d*){0,1}",ssh_spec).groups()

        password = keyring.get_password(server,username)
        if not password:
            return None

        ssh_client = SSHClient()
        ssh_client.load_system_host_keys()
        ssh_client.connect( hostname=server, port=int(port if port else 22), username=username, password=password)

        if self.setup( ssh_client ):
            session,stdin,stdout,stderror = self.start_command(ssh_client,"~/.local/bin/dashboard --server")
            connection = Connection(self,ssh_client,session,stdin,stdout,stderror)
            connection.open(client)
            self.connections[ssh_spec] = connection
            return connection

        raise Exception("Setup of remote dashboard failed")

    @sync_manager
    def start_command(self, ssh_client, command ):
        """ start a command and return a tuple with (channel,stdin,stdout,stderr) for running process """
        transport = ssh_client.get_transport()
        session = transport.open_session()
        session.exec_command(command)
        stdout = session.makefile("r",1)
        stderr = session.makefile_stderr("r",1)
        stdin = session.makefile_stdin("w",1)
        return (session,stdin,stdout,stderr)

    @sync_manager
    def run_command(self, ssh_client, command ):
        """ run a command wait for it to exit and return the output (retcode,stdout_str,stderr_str) """
        session,stdin,stdout,stderr = self.start_command(ssh_client,command)

        stderr_output = StringIO()
        stdout_output = StringIO()
        while not session.exit_status_ready():
            stdout_output.write(stdout.readline())
            stderr_output.write(stderr.readline())
        stdout_output.write("".join(stdout.readlines()))
        stderr_output.write("".join(stderr.readlines()))
        exit_status = session.recv_exit_status()
        return (exit_status,stdout_output.getvalue(),stderr_output.getvalue())

    @sync_manager
    def setup(self, ssh_client ):
        """ check to see that dashboard is installed and install it if needed """
        exit_status,stdout_str,stderr_str = self.run_command(ssh_client,"~/.local/bin/dashboard --version")
        stdout_str = stdout_str.strip()
        if stdout_str.startswith("dashboard version"):
            if stdout_str.endswith(__version__):
                return True

        exit_status,stdout_str,stderr_str = self.run_command(ssh_client,'python3 -m pip install --upgrade "terminal-dashboard==%s"'%(__version__))
        if exit_status:
            raise Exception(exit_status,stdout_str,stderr_str)

        return True


_connection_manager = None
def get_connection_manager():
    """ return the connection manager create one if it doesn't exit """
    global _connection_manager
    if not _connection_manager:
        _connection_manager = ConnectionManager()
    return _connection_manager

def shutdown_connection_manager():
    """ shut down the connection manager if it was ever started """
    global _connection_manager
    if _connection_manager:
        _connection_manager.shutdown()
        _connection_manager = None

class RemoteDataTable( DataTable ):
    def __init__(self,ssh_spec=None,table_def=None,name=None,refresh_minutes=1):
        """ accepts an ssh_spec to connect to of the form ssh://username@server_name:port_number, a json string with the definition for the remote table, the local name for this table, and the number of minutes for refresh """
        DataTable.__init__(self,None,name,refresh_minutes)
        self.ssh_spec = ssh_spec
        self.table_def = table_def
        self.connection = None
        self.refresh()

    @synchronized
    def refresh(self):
        """ create a connection to the remote dashboard table server and refresh our internal state """

        if not self.connection:
            cm = get_connection_manager()
            connection = cm.connect(self.ssh_spec,self)
            if not connection:
                return
            self.connection = connection
            response = self.connection.table(json.dumps(self.table_def))
            if not response.startswith("loaded:%s"%self.table_def["name"]):
                return

        table_data = self.connection.get(self.table_def["name"])
        name,json_blob = table_data.split(":",1)
        dt = from_json(StringIO(json_blob))

        rows,cols = dt.get_bounds()
        for idx in range(cols):
            self.replace_column(idx,dt.get_column(idx))

        self.changed()
        DataTable.refresh(self)



# Copyright 2017 James P Goodwin data table package to manage sparse columnar data
""" module that implement a data table package to manage sparse columnar data window """
import sys
import os
from datetime import datetime
import threading
import time
import csv
import json
from functools import wraps

string_type = '_string'
float_type = '_float'
int_type = '_int'
date_type = '_date'
blank_type = '_blank'

def format_string( s ):
    return str(s)

def format_date( d ):
    return d.strftime("%m/%d/%y %H:%M")

def format_float( d ):
    if d >= 1000 and d < 1000000:
        return "%.0fK"%(d/1000)
    elif d >= 1000000 and d < 1000000000:
        return "%.0fM"%(d/1000000)
    elif d >= 1000000000:
        return "%.0fG"%(d/1000000000)
    else:
        return "%.2f"%d

def format_int( d ):
    if d >= 1000 and d < 1000000:
        return "%dK"%(d//1000)
    elif d >= 1000000 and d < 1000000000:
        return "%dM"%(d//1000000)
    elif d >= 1000000000:
        return "%dG"%(d//1000000000)
    else:
        return "%d"%d

class Cell(object):
    def __init__(self,type,value,format):
        self.type = type
        self.value = value
        self.format = format

    def __str__(self):
        return self.format(self.value)

    def get_type(self):
        return self.type

    def get_value(self):
        return self.value

    def put_value(self,value):
        self.value = value

    def get_float_value(self):
        if self.type in [float_type,int_type]:
            return float(self.value)
        elif self.type == date_type:
            return self.value.timestamp()
        else:
            return 0.0

    def get_format(self):
        return self.format

    def set_format(self,format):
        self.format = format

blank_cell = Cell(blank_type,None,lambda x: "")

class ColumnIterator(object):
    def __init__(self,column):
        self.column = column
        self.idx = 0
        self.limit = column.size()

    def __iter__(self):
        return self

    def __next__(self):
        if self.idx >= self.limit:
            raise StopIteration
        ret = self.column.get(self.idx)
        self.idx += 1
        return ret


class Column(object):
    def __init__(self,values=None,idx=0,name=None,table=None):
        """ accept a list of Cell objects, a column index, and a column name, and a table to be a part of """
        self.values = values if values else []
        self.idx = 0
        self.name = name
        self.table = table

    def size(self):
        """ get the size of this column """
        return len(self.values)

    def delete(self,idx):
        if idx < len(self.values):
            del self.values[idx]

    def ins(self,idx,value):
        if idx < len(self.values):
            self.values.insert(idx,value)
        else:
            self.put(idx,value)

    def get(self,idx):
        """ get the cell at index idx in column """
        if idx < len(self.values):
            return self.values[idx]
        else:
            return blank_cell

    def put(self,idx,value):
        """ put a Cell value at index idx in column """
        while idx >= len(self.values):
            self.values.append(blank_cell)
        self.values[idx] = value

    def get_name(self):
        return self.name

    def set_name(self,name):
        self.name = name

    def get_idx(self):
        return self.idx

    def set_idx(self,idx):
        self.idx = idx

    def get_table(self):
        return self.table

    def set_table(self,table):
        self.table = table

blank_column = Column()

def synchronized(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        with self.refresh_lock:
            return method(self, *args, **kwargs)
    return wrapper

#   format of data table as json
#   {
#       "name" : name of the table,
#       "refresh_minutes" : refresh interval in minutes,
#       "columns" : [ array of column structures
#           {
#           "name": column name
#           "values" : [ array of cells in column
#               {
#               "type" : one of "_string","_float","_int","_date","_blank"
#               "value" : string, float, int, float for date, or "" for blank
#               },
#               ]
#           },
#           ]
#   }
def from_json( stream ):
    """ load a DataTable from a stream as JSON, return new DataTable """
    jtable = json.load(stream)
    dt = DataTable(None,jtable.get("name","JSON DataTable"),jtable.get("refresh_minutes",1))
    for c in jtable["columns"]:
        nc = Column( name = c.get("name", None) )
        for v in c["values"]:
            ct = v["type"]
            cv = v["value"]
            if ct == string_type:
                cc = Cell(string_type,cv,format_string)
            elif ct == float_type:
                cc = Cell(float_type,cv,format_float)
            elif ct == int_type:
                cc = Cell(int_type,cv,format_float)
            elif ct == date_type:
                cc = Cell(date_type,datetime.fromtimestamp(cv),format_date)
            elif ct == blank_type:
                cc = blank_cell
            nc.put(nc.size(),cc)
        dt.add_column( nc )
    return dt

def to_json( dt, stream ):
    """ write a DataTable to a stream as JSON """
    out_dict = {}
    if dt.name:
        out_dict["name"] = dt.name

    out_dict["refresh_minutes"] = dt.refresh_minutes

    columns  = []
    for idx in range(len(dt.columns)):
        dtc = dt.columns[idx]
        column = {}
        if dtc.name:
            column["name"] = dtc.name

        values = []
        for dtv in dtc.values:
            values.append( { "type":dtv.type, "value": ( dtv.value if dtv.type != date_type else dtv.get_float_value() ) } )
        column["values"] = values
        columns.append(column)

    out_dict["columns"] = columns
    json.dump(out_dict,stream)

# csv representation of a DataTable
# heading row at the top
# each column of the form: table name_column name_type names cannot contain '_" and
# types will be used to load cells can't have mixed cell types in a column
def from_csv( stream, name=None, field_map=None ):
    """ load a DataTable from a stream as CSV, return new DataTable, you can provide an override to the default parsing to provide a name and a field_map which is a list of tuples CSV_column_name,DataTable_column_name,DataTable_type it will only load columns in the column map """
    dt = None
    dr = csv.DictReader(stream)
    for drr in dr:
        for drc in drr:
            parts = drc.split("_",2)
            if not dt:
                if name:
                    dt = DataTable(name=name)
                elif not field_map and len(parts) == 3:
                    dt = DataTable(name=parts[0])
                else:
                    dt = DataTable()

            dtc = None
            dtt = None
            if field_map:
                for fm in field_map:
                    if drc == fm[0]:
                        dtc = fm[1]
                        dtt = fm[2]
                        break
            else:
                if len(parts) == 3:
                    dtc = parts[1]
                    dtt = "_"+parts[2]

            if dtc and dtt:
                if not dt.has_column(dtc):
                    dt.add_column(Column(name=dtc))
                dtcc = dt.get_column(dtc)

                drv = drr[drc]
                if drv and dtt == string_type:
                    cc = Cell(string_type,drv,format_string)
                elif drv and dtt == float_type:
                    cc = Cell(float_type,float(drv),format_float)
                elif drv and dtt == int_type:
                    cc = Cell(int_type,int(drv),format_float)
                elif drv and dtt == date_type:
                    cc = Cell(date_type,datetime.fromtimestamp(float(drv)),format_date)
                elif not drv or dtt == blank_type:
                    cc = blank_cell
                dtcc.put(dtcc.size(),cc)
    return dt

def to_csv( dt, stream ):
    """ write a DataTable to a stream as CSV, see standard format in comments above, type for column is based on the zeroth cell """
    field_names = []
    idx = 0
    max_idx = 0
    for c in dt.columns:
        type = blank_type
        for tidx in range(c.size()):
            if c.get(tidx).type != blank_type:
                type = c.get(tidx).type
                break
        field_names.append((dt.name if dt.name else "DataTable")+"_"+(c.name if c.name else "Column %d"%idx)+type)
        idx += 1
        if c.size() > max_idx:
            max_idx = c.size()
    wcsv = csv.DictWriter(stream,field_names)
    for wcridx in range(max_idx):
        wcr = {}
        for idx in range(len(dt.columns)):
            cell = dt.columns[idx].get(wcridx)
            wcr[field_names[idx]] = (cell.get_value() if cell.type != date_type else cell.get_float_value())
        if wcridx == 0:
            wcsv.writeheader()
        wcsv.writerow(wcr)

class DataTable(object):
    def __init__(self,columns=None,name=None,refresh_minutes=10):
        """ accepts a list of columns and a name for the table """
        self.listeners = []
        self.columns = []
        self.name = name
        self.cnames = {}
        self.refresh_lock = threading.RLock()
        self.refresh_minutes = refresh_minutes
        self.refresh_thread = None
        self.refresh_thread_stop = False
        self.refresh_timestamp = None
        if columns:
            for c in columns:
                self.add_column(c)

    def get_refresh_timestamp( self ):
        """ get the time that the table was last refreshed """
        return self.refresh_timestamp

    def acquire_refresh_lock(self):
        """ acquire the refresh lock before reading/writing the table state """
        self.refresh_lock.acquire()

    def release_refresh_lock(self):
        """ release the refresh lock after reading/writing the table state """
        self.refresh_lock.release()

    def start_refresh( self ):
        """ Start the background refresh thread """
        self.stop_refresh()
        self.refresh_thread = threading.Thread(target=self.perform_refresh)
        self.refresh_thread.start()

    def perform_refresh( self ):
        """ Thread worker that sleeps and refreshes the data on a schedule """
        start_time = time.time()
        while not self.refresh_thread_stop:
            if time.time() - start_time >= self.refresh_minutes*60.0:
                self.refresh()
                start_time = time.time()
            time.sleep(0)

    def stop_refresh( self ):
        """ Stop the background refresh thread """
        self.refresh_thread_stop = True
        if self.refresh_thread and self.refresh_thread.is_alive():
            self.refresh_thread.join()
        self.refresh_thread = None
        self.refresh_thread_stop = False

    def listen(self,listen_func):
        """ register for notifications when a change event is raised on this table """
        self.listeners.append(listen_func)

    def unlisten(self,listen_func):
        """ unregister for notifications when a change event is raised on this table """
        self.listeners.remove(listen_func)

    def changed(self):
        """ notify listeners that this table has been changed """
        for f in self.listeners:
            f(self)

    @synchronized
    def get_bounds(self):
        """ return a tuple (rows,cols) where rows is the maximum number of rows and cols is the maximum number of cols """
        cols = len(self.columns)
        rows = -1
        for c in self.columns:
            size = c.size()
            if rows < 0 or size > rows:
                rows = size
        return (rows,cols)

    def get_name(self):
        """ return the name of the table """
        return self.name

    @synchronized
    def get_names(self):
        """ return a list of the names of the columns in order"""
        return [c.get_name() for c in self.columns]

    @synchronized
    def get_columns(self):
        """ return the list of columns """
        return self.columns

    @synchronized
    def add_column(self,column):
        idx = len(self.columns)
        column.set_idx(idx)
        if not column.get_name():
            column.set_name("%s_%d"%(self.name,idx))
        self.columns.append(column)
        self.cnames[column.get_name()] = column
        column.set_table(self)

    @synchronized
    def insert_column(self,idx,column):
        while idx > len(self.columns):
            self.add_column(blank_column)
        if idx == len(self.columns):
            self.add_column(column)
        else:
            if not column.get_name():
                column.set_name("%s_%d"%(self.name,idx))
            self.columns.insert(idx,column)
            self.cnames[column.get_name()] = column
            column.set_table(self)
            while idx < len(self.columns):
                if column.get_name() == "%s_%d"%(self.name,idx-1):
                    column.set_name("%s_%d"%(self.name,idx))
                    self.cnames[column.get_name()] = column
                self.columns[idx].set_idx(idx)
                idx += 1

    @synchronized
    def replace_column(self,idx,column):
        column.set_idx(idx)
        if not column.get_name():
            column.set_name("%s_%d"%(self.name,idx))
        if idx == len(self.columns):
            self.columns.append(column)
        else:
            self.columns[idx] = column
        self.cnames[column.get_name()] = column
        column.set_table(self)

    @synchronized
    def map_column(self, reference ):
        if type(reference) == str or type(reference) == str:
            return self.cnames[reference].get_idx()
        elif type(reference) == int:
            return reference
        else:
            raise TypeError("wrong type in mapping")

    @synchronized
    def has_column(self, reference ):
        if type(reference) == str or type(reference) == str:
            return reference in self.cnames
        elif type(reference) == int:
            return idx < len(self.columns)
        else:
            return False

    @synchronized
    def get_column(self, reference):
        return self.columns[self.map_column(reference)]

    @synchronized
    def get(self, row, reference ):
        return self.columns[self.map_column(reference)].get(row)

    @synchronized
    def put(self, row, reference, value):
        self.columns[self.map_column(reference)].put(row,value)

    @synchronized
    def refresh(self):
        """ base class method for forcing a refresh on a table """
        self.refresh_timestamp = time.time()



# Copyright 2020 James P Goodwin data table package to manage sparse columnar data
""" module that aggregates data from the syslog and provides a set of data tables """
import locale
locale.setlocale(locale.LC_ALL,'')
import sys
import os
import glob
import gzip
import re
import psutil
from datetime import datetime,timedelta
from data_sources.data_table import DataTable,Column,Cell,blank_type,string_type,float_type,int_type,date_type,format_string,format_float,format_date,format_int,synchronized

class AverageCell(Cell):
    def __init__(self,type,value,format):
        self.type = type
        self.value = value
        self.format = format
        self.values = [ value ]

    def put_value(self,value):
        self.values.append(value)
        self.value = sum(self.values)/len(self.values)

class ProcDataTable( DataTable ):
    """ class that collects a time based aggregation of data from system process information into a data_table """
    def __init__(self,num_hours=24,bucket_hours=1,refresh_minutes=10):
        """ Initialize the ProcDataTable to collect system process information, a timespan to aggregate for, aggregation bucket in hours, a refresh interval for updating in minutes """
        self.num_hours = num_hours
        self.bucket_hours = bucket_hours
        DataTable.__init__(self,None,
            "Proc Data: for the last %d hours in %d hour buckets, refreshed every %d minutes"%(
            self.num_hours,
            self.bucket_hours,
            refresh_minutes),refresh_minutes)
        self.refresh()

    @synchronized
    def refresh( self ):
        """ refresh or rebuild tables """

        current_time = datetime.now()

        column_names = [ "Time Stamps", "CPU Percent", "Load Avg",
            "Total Virtual Memory", "Available Virtual Memory",
            "Filesystem Percent Full", "Filesystem Read Bytes", "Filesystem Write Bytes",
            "Network Sent Bytes","Network Received Bytes","Network Connections" ]

        def bucket_idx( timestamp, column ):
            for idx in range(column.size()):
                if column.get(idx).get_value() >= timestamp:
                    return idx
            else:
                return -1

        def append_bucket( timestamp, column ):
            column.put(column.size(),Cell(date_type,timestamp+timedelta( hours=self.bucket_hours),format_date))
            if column.size() > self.num_hours/self.bucket_hours:
                for cn in column_names:
                    self.get_column(cn).delete(0)
            return column.size()-1

        def add_average( column_name, idx, value ):
            column = self.get_column(column_name)
            if not column.size() or idx >= column.size():
                column.put(idx, AverageCell(float_type,value,format_float))
            else:
                column.get(idx).put_value(value)

        bidx = 0
        for cn in column_names:
            if not self.has_column(cn):
                self.add_column(Column(name=cn))
            if cn == "Time Stamps":
                bidx = bucket_idx( current_time, self.get_column(cn))
                if bidx < 0:
                    bidx = append_bucket( current_time, self.get_column(cn))
            elif cn == "CPU Percent":
                add_average(cn,bidx,psutil.cpu_percent())
            elif cn == "Load Avg":
                add_average(cn,bidx,psutil.getloadavg()[2])
            elif cn == "Total Virtual Memory":
                add_average(cn,bidx,psutil.virtual_memory().total)
            elif cn == "Available Virtual Memory":
                add_average(cn,bidx,psutil.virtual_memory().available)
            elif cn == "Filesystem Percent Full":
                add_average(cn,bidx,psutil.disk_usage("/").percent)
            elif cn == "Filesystem Read Bytes":
                add_average(cn,bidx,psutil.disk_io_counters().read_bytes)
            elif cn == "Filesystem Write Bytes":
                add_average(cn,bidx,psutil.disk_io_counters().write_bytes)
            elif cn == "Network Sent Bytes":
                add_average(cn,bidx,psutil.net_io_counters().bytes_sent)
            elif cn == "Network Recieved Bytes":
                add_average(cn,bidx,psutil.net_io_counters().bytes_recv)
            elif cn == "Network Connections":
                add_average(cn,bidx,float(len(psutil.net_connections())))
        self.changed()

        DataTable.refresh(self)



# Copyright 2017 James P Goodwin a package to draw graphs of data in a terminal window
""" module that implements graphs of data rendered in a terminal window """
import locale
locale.setlocale(locale.LC_ALL,"")
import curses
import curses.ascii
import sys
import os
import math
import time
from datetime import datetime
from char_draw import canvas
from char_draw import graph
from data_sources import data_table,syslog_data

class Panel:
    def __init__(self, y=-1,x=-1,height=-1,width=-1,graphs=None,parent=None):
        """ A panel is a container for a collection of graphs they fit within it's bounds """
        self.x = x
        self.y = y
        self.height = height
        self.width = width
        self.graphs = graphs if graphs else []
        self.graph_layout = []
        self.parent = None
        self.pad = None
        if parent:
            self.set_parent( parent )

    def set_layout( self, y, x, height, width ):
        """ set the current layout of this panel """
        self.y = y
        self.x = x
        self.height = height
        self.width = width
        self.layout_graphs()

    def get_layout( self ):
        """ return the panel layout """
        return (self.y,self.x,self.height,self.width)

    def get_pad( self ):
        """ return this panel's pad just a passthru to the parent's pad """
        return self.pad

    def set_parent( self, parent ):
        """ set this panel to have this parent """
        self.pad = parent.get_pad()
        self.parent = parent
        self.layout_graphs()

    def add_graph( self, graph ):
        """ add a graph to this panel """
        self.graphs.append( graph )
        self.layout_graphs()

    def layout_graphs( self ):
        """ compute a rectangular layout of graphs within the panel and assign them those locations """
        self.graph_layout = []
        if self.x >= 0: # if the panel has been sized
            ngraphs = len(self.graphs)
            best_fit = None
            min_dxy = None
            for cols in range(1,ngraphs+1):
                for rows in range(1,ngraphs+1):
                    if rows*cols >= ngraphs:
                        dx = self.width//cols
                        dy = self.height//rows
                        dxy = max(dx,dy)-min(dx,dy)
                        if min_dxy == None or dxy < min_dxy:
                            min_dxy = dxy
                            best_fit = (rows,cols,dy,dx)
            rows,cols,dy,dx = best_fit
            x = 0
            y = 0
            for g in self.graphs:
                self.graph_layout.append((y,x,dy,dx))
                pad = self.get_pad()
                if pad:
                    g_pad = pad.subpad(dy,dx,y,x)
                    g_canvas = canvas.Canvas(g_pad)
                    g.set_canvas(g_canvas)
                x += dx
                if x >= self.width:
                    x = 0
                    y += dy
        return self.graph_layout

    def get_graph_layout( self ):
        """ return a list of tuples (graph,y,x,height,width) for the current layout """
        if not self.graph_layout:
            self.layout_graphs()
        ret_layout = []
        idx = 0
        for y,x,dy,dx in self.graph_layout:
            ret_layout.append((self.graphs[idx],self.y+y,self.x+x,dy,dx))
            idx += 1
        return ret_layout

    def redraw( self ):
        """ redraw the panel and all it's graphs """
        if not self.graph_layout:
            self.layout_graphs()
        if self.get_pad():
            for g in self.graphs:
                if g.is_modified():
                    g.render()

class Page:
    def __init__(self, window, height=-1, width=-1 ):
        """ one page of a dashboard has the parent window and it's height and width """
        self.window = window

        if height <= 0 or width <= 0:
            height,width = window.getmaxyx()
            height = height - 1

        self.height = height
        self.width = width
        self.pad = curses.newpad(height,width)
        self.position = (0,0)
        self.panels = []
        self.panel_layout = []
        self.current_panel = 0

    def get_pad( self ):
        """ return  your pad """
        return self.pad

    def add_panel( self, panel ):
        """ add a panel into the page """
        self.panels.append(panel)
        panel.set_parent(self)
        self.layout_panels()

    def move( self, row, col ):
        """ move the current scroll position to a new location """
        if row < self.height and row >= 0 and col < self.width and col >= 0:
            self.position = (row,col)
        return self.position

    def get_position( self ):
        """ get the current position """
        return self.position

    def get_size( self ):
        """ return the height and width of this page """
        return (self.height,self.width)

    def get_current_panel( self ):
        """ return the current panel """
        panel = None
        if self.panels:
            panel = self.panels[self.current_panel]
        return panel

    def next_panel( self ):
        """ move to the next panel and return that panel or None if we're at the end"""
        panel = None
        if self.panels:
            if self.current_panel+1 < len(self.panels):
                self.current_panel = self.current_panel + 1
                panel = self.panels[self.current_panel]
        return panel

    def prev_panel( self ):
        """ move to the previous panel return that panel or None if we're at the start """
        panel = None
        if self.panels:
            if self.current_panel > 0:
                self.current_panel = self.current_panel -1
                panel = self.panels[self.current_panel]
        return panel

    def first_panel( self ):
        """ move to the first panel and return that panel """
        panel = None
        if self.panels:
            self.current_panel = 0
            panel = self.panels[self.current_panel]
        return panel

    def last_panel( self ):
        """ move to the last panel and return that panel """
        panel = None
        if self.panels:
            self.current_panel = len(self.panels)-1
            panel = self.panels[self.current_panel]
        return panel

    def redraw( self ):
        """ redraw all of the panels """
        if not self.get_panel_layout():
            self.layout_panels()

        for p in self.panels:
            p.redraw()

    def layout_panels( self ):
        """ compute a rectangular layout of panels within the page and assign them those locations """
        self.panel_layout = []

        is_dynamic = False
        for g in self.panels:
            y,x,dy,dx = g.get_layout()
            if x < 0:
                is_dynamic = True

        if is_dynamic:
            npanels = len(self.panels)
            best_fit = None
            min_dxy = None
            for cols in range(1,npanels+1):
                for rows in range(1,npanels+1):
                    if rows*cols >= npanels:
                        dx = self.width//cols
                        dy = self.height//rows
                        dxy = max(dx,dy)-min(dx,dy)
                        if min_dxy == None or dxy < min_dxy:
                            min_dxy = dxy
                            best_fit = (rows,cols,dy,dx)

            rows,cols,dy,dx = best_fit
            x = 0
            y = 0

            for g in self.panels:
                self.panel_layout.append((g,y,x,dy,dx))
                g.set_layout(y,x,dy,dx)
                x += dx
                if x >= self.width:
                    x = 0
                    y += dy
        else:
            for g in self.panels:
                y,x,dy,dx = g.get_layout()
                self.panel_layout.append((g,y,x,dy,dx))

        return self.panel_layout

    def get_panel_layout( self ):
        """ return the layout for the panels, tuples are panel,y,x,height,width """
        return self.panel_layout

    def refresh( self ):
        """ refresh this page showing the current position """
        w_ymax,w_xmax = self.window.getmaxyx()
        w_ymax -= 1
        if w_ymax >= self.height and w_xmax >= self.width:
            self.pad.refresh(0,0,1,0,self.height,self.width)
        else:
            p_y,p_x = self.position
            h = min(w_ymax,self.height - p_y)
            w = min(w_xmax,self.width - p_x)
            self.pad.refresh(p_y,p_x,1,0,h-1,w-1)
            if w < w_xmax:
                for y in range(0,h):
                    try:
                        self.window.addstr(y,w,' '*(w_xmax-w))
                    except:
                        pass
            while h < w_ymax:
                try:
                    self.window.addstr(h,0,' '*(w_xmax-1))
                except:
                    pass
                h += 1

class Dashboard:
    def __init__(self,window,pages = None, auto_tour_delay = 0):
        """ Dashboard driver that holds an array of pages, auto_tour_delay > 0 will tab around all of the graphs with this delay in seconds between each """
        self.window = window
        self.pages = pages if pages else []
        self.current_page = 0
        self.current_panel = None
        self.current_graph = 0
        self.current_graph_pos = None
        self.auto_tour_delay = auto_tour_delay

    def add_page(self, page ):
        """ add a page to this dashboard """
        self.pages.append(page)

    def get_window( self ):
        """ Get this dashboard's window """
        return self.window

    def get_current_page( self ):
        """ Get the current page """
        page = None
        if self.pages:
            page = self.pages[self.current_page]
        return page


    def reset_position( self ):
        """ Reset the tour position """
        page = self.get_current_page()
        if page:
            self.current_panel = page.first_panel()
            self.current_graph = 0
            page.move(0,0)
            self.current_graph_pos = None

    def next_page( self ):
        """ Goto the next page and return it or return None if at the end """
        page = None
        if self.pages:
            if self.current_page+1 < len(self.pages):
                self.current_page += 1
                page = self.pages[self.current_page]
                self.reset_position()
        return page

    def prev_page( self ):
        """ goto the previous page and return it or return None if at the beginning """
        page = None
        if self.pages:
            if self.current_page > 0:
                self.current_page -= 1
                page = self.pages[self.current_page]
                self.reset_position()
        return page

    def first_page( self ):
        """ goto the first page and return it or return None if no pages """
        page = None
        if self.pages:
            self.current_page = 0
            page = self.pages[self.current_page]
            self.reset_position()
        return page

    def last_page( self ):
        """ goto the last page and return it or return None if no pages """
        page = None
        if self.pages:
            self.current_page = len(self.pages)-1
            page = self.pages[self.current_page]
            self.reset_position()
        return page

    def get_current_panel( self ):
        """ get the current_panel in the current_page or None if there are none """
        panel = None
        page = self.get_current_page()
        if page:
            panel = page.get_current_panel()
        return panel

    def get_current_graph( self ):
        """ get the current graph tuple in the current_page in the current panel, None if there is none """
        graph = None
        panel = self.get_current_panel()
        if panel:
            graphs = panel.get_graph_layout()
            if graphs:
                if self.current_graph >= len(graphs):
                    self.current_graph = 0
                graph = graphs[self.current_graph]
        return graph

    def next_graph( self ):
        """ step forward to the next graph,panel,page return the graph tuple for the current graph, return None at the end """
        graph = None
        panel = self.get_current_panel()
        page = self.get_current_page()
        if panel:
            graphs = panel.get_graph_layout()
            while panel and graphs and not graph:
                if self.current_graph+1 < len(graphs):
                    self.current_graph += 1
                    graph = graphs[self.current_graph]
                else:
                    page = self.get_current_page()
                    panel = page.next_panel()
                    self.current_graph = 0
                    if not panel:
                        page = self.next_page()
                        if page:
                            graph = self.get_current_graph()
                        else:
                            return None
                    else:
                        graphs = panel.get_graph_layout()
                        graph = graphs[self.current_graph]
            if graph:
                page.move(graph[1],graph[2])

        return graph


    def main( self ):
        """ main input and redraw loop for the dashboard """
        self.window.nodelay(1)
        self.window.notimeout(0)
        self.window.timeout(0)
        self.window.keypad(1)
        curses.curs_set(0)
        curses.raw()
        curses.mousemask( curses.ALL_MOUSE_EVENTS | curses.REPORT_MOUSE_POSITION )

        start_time = time.time()
        positions = []
        zoomed = False
        zoomed_canvas = None
        zoomed_graph = None
        zoomed_restore_canvas = None
        status = ""
        while True:
            ch = self.window.getch()

            if not zoomed:
                cur_graph = self.get_current_graph()
                cur_graph[0].set_focus(True)

                if ch > -1:
                    page = self.get_current_page()
                    py,px = page.get_position()

                    if ch == 27: # esc key
                        return 0
                    elif ch == 9: # tab key
                        self.next_graph()
                    elif ch == curses.KEY_NPAGE:
                        self.next_page()
                    elif ch == curses.KEY_PPAGE:
                        self.prev_page()
                    elif ch == curses.KEY_HOME:
                        self.first_page()
                    elif ch == curses.KEY_END:
                        self.last_page()
                    elif ch == curses.KEY_RIGHT:
                        page.move(py,px+1)
                    elif ch == curses.KEY_LEFT:
                        page.move(py,px-1)
                    elif ch == curses.KEY_UP:
                        page.move(py-1,px)
                    elif ch == curses.KEY_DOWN:
                        page.move(py+1,px)
                    elif ch == curses.KEY_F5:
                        for p in self.pages:
                            for pp in p.get_panel_layout():
                                for gg in pp[0].get_graph_layout():
                                    gg[0].refresh_data()
                    elif ch == curses.KEY_ENTER or ch == 10: # ENTER KEY
                        zoomed = True
                        zoomed_graph = self.get_current_graph()
                        zoomed_canvas = canvas.Canvas(self.window)
                        zoomed_restore_canvas = zoomed_graph[0].get_canvas()
                        zoomed_graph[0].set_canvas(zoomed_canvas)
                        self.window.clear()
                        continue
                    elif ch == curses.KEY_MOUSE:
                        mid, mx, my, mz, mtype = curses.getmouse()
                        max_y,max_x = self.window.getmaxyx()
                        if mtype & curses.BUTTON1_CLICKED:
                            if mx == 0 and my == 0:
                                self.first_page()
                            elif mx == 0 and my > 0:
                                self.prev_page()
                            elif mx == max_x-1 and my == max_y-1:
                                self.last_page()
                            elif mx == max_x-1 and my < max_y-1:
                                self.next_page()
                    start_time = time.time()
                    positions = []

                if self.auto_tour_delay:
                    if time.time() - start_time > (self.auto_tour_delay/4): # 4 views per graph
                        if not positions:
                            g = self.next_graph()
                            if not g:
                                self.first_page()
                                g = self.get_current_graph()
                            positions = [(g[1],g[2]),(g[1]+(g[3]//2),g[2]),(g[1]+(g[3]//2),g[2]+(g[4]//2)),(g[1],g[2]+(g[4]//2))]
                        py,px = positions.pop(0)
                        self.get_current_page().move(py,px)
                        start_time = time.time()

                if cur_graph[0] != self.get_current_graph()[0]:
                    cur_graph[0].set_focus(False)

                page = self.get_current_page()
                if page:
                    page.redraw()
                    page.refresh()

            if zoomed:
                if ch > -1:
                    if ch == 27: # esc key
                        zoomed_graph[0].set_canvas(zoomed_restore_canvas)
                        zoomed = False
                        zoomed_graph = None
                        zoomed_canvas = None
                        zoomed_restore_canvas = None

                if zoomed_graph and zoomed_graph[0].is_modified():
                    zoomed_graph[0].render()
                    self.window.refresh()

            latest_refresh = 0
            for p in self.pages:
                for pp in p.get_panel_layout():
                    for gg in pp[0].get_graph_layout():
                        latest_refresh = max(latest_refresh,gg[0].get_data().get_refresh_timestamp())

            self.window.addstr(0,0," "*len(status),curses.color_pair(1)|curses.A_REVERSE)
            status = "Page %d of %d Last Update %s%s"%(self.current_page+1,len(self.pages),datetime.fromtimestamp(latest_refresh).strftime("%A, %d. %B %Y %I:%M%p")," ZOOMED (Press Esc to Exit)" if zoomed else "")
            self.window.addstr(0,0,status,curses.color_pair(1)|curses.A_REVERSE)



def main(stdscr):
    """ test driver for the dashboard """

    sd = syslog_data.SyslogDataTable(refresh_minutes=1)
    sd.start_refresh()
    d = Dashboard(stdscr,auto_tour_delay = 5)
    p = Page(stdscr)
    height, width = p.get_size()
    pp = Panel(0,0,height,width)
    pp.add_graph(graph.LineGraph(sd,"Time Stamps",["Errors by Time"],None,None))
    pp.add_graph(graph.LineGraph(sd,"Time Stamps",["Messages by Time"],None,None))
    p.add_panel(pp)
    d.add_page(p)
    p = Page(stdscr)
    pp = Panel(0,0,height,width)
    pp.add_graph(graph.LineGraph(sd,"Time Stamps",["Warnings by Time"],None,None))
    pp.add_graph(graph.BarGraph(sd,"Services",["Messages by Service"],None,None,5))
    p.add_panel(pp)
    d.add_page(p)
    d.main()
    sd.stop_refresh()

    return 0

if __name__ == '__main__':
    curses.wrapper(main)



# Copyright 2017 James P Goodwin chardraw unicode curses based graphics package
""" module that implements a graphics package using block graphics on curses window """
import sys
import os
from char_draw import canvas
import curses


class Bbox:
    """ Bounding box object """
    def __init__(self, x0=0,y0=0,x1=0,y1=0):
        """ constructor, takex the bounding points of the bounding box """
        self.x0 = x0 if x0 < x1 else x1
        self.y0 = y0 if y0 < y1 else y1
        self.x1 = x1 if x1 > x0 else x0
        self.y1 = y1 if y1 > y0 else y0

    def get_size(self):
        """ return the height and width of the bounding box """
        return (self.x1-self.x0),(self.y1-self.y0)

    def get_location(self):
        """ return the location of the upper left corner of the bounding box """
        return self.x0,self.y0

    def union(self,bbox):
        """ constructs the bounding box that is the union of this bounding box and the input bbox """
        ret = Bbox( self.x0, self.y0, self.x1, self.y1 )
        ret.x0 = bbox.x0 if bbox.x0 < self.x0 else self.x0
        ret.y0 = bbox.y0 if bbox.y0 < self.y0 else self.y0
        ret.x1 = bbox.x1 if bbox.x1 > self.x1 else self.x1
        ret.y1 = bbox.y1 if bbox.y1 > self.y1 else self.y1
        return (ret)

    def __repr__(self):
        return "(%10.2f,%10.2f,%10.2f,%10.2f)"%(self.x0,self.y0,self.x1,self.y1)


class DisplayList:
    """ Base object for all display list objects """
    def __init__(self, parent = None, children = None, canvas=None):
        """ constructor, takes a parent object to be contained within and a list of children object to contain """
        self.parent = parent
        if children:
            self.children = children
        else:
            self.children = []
        # modified flag indicates if the object has been changed
        self.modified = True
        # initialize bbox
        self.bbox = None
        # set up canvas
        self.canvas = None
        self.xform = [1.0,1.0]
        self.focus = False
        if canvas:
            self.set_canvas(canvas)
        elif parent:
            self.set_canvas(parent.get_canvas())

    def set_focus( self, state  ):
        """ set focus to True or False to tell graph to render in focus mode """
        self.focus = state
        self.modified = True

    def is_focus( self ):
        """ return the current focus state """
        return self.focus

    def set_parent(self,parent):
        """ set the parent of this object """
        self.parent = parent
        self.set_canvas(parent.get_canvas())

    def get_parent(self):
        """ get the parent of this object """
        return(self.parent)

    def get_canvas(self):
        """ get the canvas for this display list """
        return self.canvas

    def set_canvas(self,canvas):
        """ set the canvas and the scaling transformation """
        self.xform = [1.0,1.0]
        self.canvas = canvas
        self.modified = True
        for c in self.children:
            c.set_canvas(canvas)

    def get_bbox(self):
        """ computes the bounding box of the display list object """
        ret = self.bbox
        if self.modified:
            ret = self.bbox = None
            for c in self.children:
                if not ret:
                    ret = c.get_bbox()
                else:
                    ret = ret.union(c.get_bbox())
            self.bbox = ret
            self.modified = False
        return (ret)

    def render(self):
        """ draw the object on the suppied canvas """
        for c in self.children:
            c.render()

    def transform(self, x, y ):
        """ apply current transform to point """
        return (x*self.xform[0], y*self.xform[1])

    def bounds(self):
        """ return the bounds of the current drawing surface """
        max_x,max_y = self.canvas.get_maxxy()
        return (0.0,0.0,max_x/self.xform[0],max_y/self.xform[1])

    def pick(self, row, col):
        """ return the topmost opject at row,col for mouse selection """
        for c in self.children:
            o = c.pick(row,col)
            if o:
                return o
        b = self.get_bbox()
        ulrow,ulcol = self.canvas.to_rowcol(*self.transform(b.x0,b.y0))
        lrrow,lrcol = self.canvas.to_rowcol(*self.transform(b.x1,b.y1))
        if row >= ulrow and row <= lrrow and col >= ulcol and col <= lrcol:
            return self
        else:
            return None

    def invalidate(self):
        """ mark this object and it's parent as invalidated """
        self.modified = True
        if self.parent:
            self.parent.invalidate()

    def is_modified(self):
        """ does this display list contain modified objects """
        modified = self.modified
        for c in self.children:
            modified = modified or c.is_modified()
        return modified

    def handle(self, event ):
        """ handle an input event, return None if handled, event if not """
        return event

    def add_child(self, child ):
        """ append a child to the list of children of this object """
        self.children.append(child)
        child.set_parent(self)
        child.set_canvas(self.get_canvas())
        self.invalidate()

    def remove_child(self, child ):
        """ remove this child from the list of children of this object """
        self.children.remove(child)
        child.set_parent(None)

    def get_children(self):
        """ return the children of this object if any """
        return self.children

    def set_children(self, children):
        """ set the children of this object to a new list of children """
        self.children = children
        for c in children:
            c.set_parent(self)
            c.set_canvas(self.get_canvas())
        self.invalidate()


class Rect(DisplayList):
    """Display list object for a rectangle"""
    def __init__(self,x0,y0,x1,y1,color,fill=False,parent=None):
        """ x0,y0 upper left corner of box, x1,y1 lower right corner of box, color is color, fill=True to fill parent is parent object """
        DisplayList.__init__(self,parent,None,None)
        self.x0 = x0
        self.y0 = y0
        self.x1 = x1
        self.y1 = y1
        self.color = color
        self.fill = fill
        self.bbox = None

    def get_points( self ):
        """ return the rectangle's raw points """
        return ((self.x0,self.y0),(self.x1,self.y1))

    def set_points( self, points ):
        """ set the rectangle's raw points """
        self.x0,self.y0 = points[0]
        self.x1,self.y1 = points[1]
        self.invalidate()

    def get_color( self ):
        """ return the color of the rectangle """
        return self.color

    def set_color( self, color):
        """ set the color of the rectangle """
        self.color = color
        self.invalidate()

    def get_fill( self ):
        """ return the fill state of the rectangle """
        return self.fill

    def set_fill( self, fill):
        """ set the filled state of the rectangle """
        self.fill = fill
        self.invalidate()

    def render(self):
        """ override rendering to draw the requested rectangle """
        x0,y0 = self.transform( self.x0,self.y0 )
        x1,y1 = self.transform( self.x1,self.y1 )
        self.canvas.rect(self.x0,self.y0,self.x1,self.y1,self.color,self.fill)

    def get_bbox(self):
        """ get the bounding box for the rectangle """
        ret = self.bbox
        if self.modified and self.canvas:
            x0,y0 = self.transform( self.x0,self.y0 )
            x1,y1 = self.transform( self.x1,self.y1 )
            self.bbox = ret = Bbox(x0,y0,x1,y1)
            self.modified = False
        return self.bbox

class Circle(DisplayList):
    """Display list object for a circle"""
    def __init__(self, x,y,radius,color,fill=False, parent=None):
        """ x,y center of circle, radius is the radius,color is the color of the circle, fill is true to fill the circle """
        DisplayList.__init__(self,parent,None,None)
        self.x = x
        self.y = y
        self.radius = radius
        self.color = color
        self.fill = fill
        self.bbox = None

    def get_center( self ):
        """ return the center of the circle """
        return (self.x, self.y)

    def set_center( self, center ):
        """ set the center of the circle """
        self.x, self.y = center
        self.invalidate()

    def get_radius( self ):
        """ return the center of the circle """
        return self.radius

    def set_radius( self, radius ):
        """ set the center of the circle """
        self.radius = radius
        self.invalidate()

    def get_color( self ):
        """ return the color of the circle """
        return self.color

    def set_color( self, color):
        """ set the color of the circle """
        self.color = color
        self.invalidate()

    def get_fill( self ):
        """ return the fill state of the circle """
        return self.fill

    def set_fill( self, fill):
        """ set the filled state of the circle """
        self.fill = fill
        self.invalidate()

    def render(self):
        """ override rendering to draw the requested circle """
        cx,cy = self.transform( self.x,self.y )
        dx,dy = self.transform( self.x+self.radius, self.y )
        sr = dx-cx
        self.canvas.circle(cx,cy,sr,self.color,self.fill)

    def get_bbox(self):
        """ compute the bbox of the circle """
        ret = self.bbox
        if self.canvas and self.modified:
            cx,cy = self.transform( self.x,self.y )
            dx,dy = self.transform( self.x+self.radius, self.y )
            sr = dx-cx
            self.bbox = ret = Bbox(self.x-sr,self.y-sr,self.x+sr,self.y+sr)
            self.modified = False
        return ret


class Arc(DisplayList):
    """Display list object for an arc"""
    def __init__(self, x,y,radius,a0,a1,color,fill=False, parent=None):
        """ x,y center of arc, angles are from a0 to a1, radius is the radius,color is the color of the arc, fill is true to fill the arc """
        DisplayList.__init__(self,parent,None,None)
        self.x = x
        self.y = y
        self.radius = radius
        self.color = color
        self.fill = fill
        self.a0 = a0
        self.a1 = a1
        self.bbox = None

    def get_angles( self ):
        """ get the start and end angles for the arc """
        return (self.a0,self.a1)

    def set_angles( self, angles ):
        """ set the start and end angles for the arc """
        self.a0, self.a1 = angles

    def get_center( self ):
        """ return the center of the arc """
        return (self.x, self.y)

    def set_center( self, center ):
        """ set the center of the arc """
        self.x, self.y = center
        self.invalidate()

    def get_radius( self ):
        """ return the center of the arc """
        return self.radius

    def set_radius( self, radius ):
        """ set the center of the arc """
        self.radius = radius
        self.invalidate()

    def get_color( self ):
        """ return the color of the arc """
        return self.color

    def set_color( self, color):
        """ set the color of the arc """
        self.color = color
        self.invalidate()

    def get_fill( self ):
        """ return the fill state of the arc """
        return self.fill

    def set_fill( self, fill):
        """ set the filled state of the arc """
        self.fill = fill
        self.invalidate()

    def render(self):
        """ override rendering to draw the requested arc """
        cx,cy = self.transform( self.x,self.y )
        dx,dy = self.transform( self.x+self.radius, self.y )
        sr = dx-cx
        self.canvas.arc(cx,cy,sr,self.a0,self.a1,self.color,self.fill)

    def get_bbox(self):
        """ get the bounding box, compute it from the arc points """
        ret = self.bbox
        if self.canvas and self.modified:
            b = Bbox(-1.0,-1.0,-1.0,-1.0)
            def update_bounds(x,y,color):
                if b.x0 < 0 or x < b.x0:
                    b.x0 = x
                if b.y0 < 0 or y < b.y0:
                    b.y0 = y
                if b.x1 < 0 or x > b.x1:
                    b.x1 = x
                if b.y1 < 0 or y > b.y1:
                    b.y1 = y
            cx,cy = self.transform( self.x,self.y )
            dx,dy = self.transform( self.x+self.radius, self.y )
            sr = dx-cx
            self.canvas.arc(cx,cy,sr,self.a0,self.a1,self.color,self.fill,update_bounds)
            self.bbox = ret = b
            self.modified = False
        return ret

class Text(DisplayList):
    """Display list object for text"""
    def __init__(self, x,y,message,color,parent=None):
        """ x,y is the position of the first character, message is the text to display, color is the color """
        DisplayList.__init__(self,parent,None,None)
        self.x = x
        self.y = y
        self.message = message
        self.color = color
        self.bbox = None

    def get_location(self):
        """ get the text location """
        return (self.x,self.y)

    def set_location(self, location):
        """ set the text location """
        self.x, self.y = location
        self.invalidate()

    def get_message(self):
        """ get the message """
        return self.message

    def set_message(self, message):
        """ set the message """
        self.message = message
        self.invalidate()

    def get_color( self ):
        """ return the color of the text """
        return self.color

    def set_color( self, color):
        """ set the color of the text """
        self.color = color
        self.invalidate()

    def render(self):
        """ override rendering to draw the requested text """
        cx,cy = self.transform( self.x,self.y )
        self.canvas.textat(cx,cy,self.color,self.message)

    def get_bbox(self):
        """compute bounding box of text"""
        ret = self.bbox
        if self.canvas and self.modified:
            cx,cy = self.transform(self.x,self.y)
            cx,cy = self.canvas.round_text_position(cx,cy)
            ty,tx = self.canvas.from_rowcol(1,len(self.message))
            max_x,max_y = self.canvas.get_maxxy()
            cx1 = cx+tx
            cy1 = cy+ty
            if cx1 > max_x:
                cx1 = max_x
            if cy1 > max_y:
                cy1 = max_y
            self.bbox = ret = Bbox(cx,cy,cx+tx,cy+ty)
            self.modified = False
        return ret

class PolyLine(DisplayList):
    """Display list object for polyline"""
    def __init__(self,points,color,parent=None):
        """ points is a list of (x,y) tuples in order to be drawn as a polyline in color """
        self.points = points
        self.color = color
        self.bbox = None
        self.transformed_points = None
        DisplayList.__init__(self,parent,None,None)

    def get_color( self ):
        """ return the color of the polyline """
        return self.color

    def set_color( self, color):
        """ set the color of the polyline """
        self.color = color
        self.invalidate()

    def get_points(self):
        """ return a tuple of x,y tuples representing this PolyLine """
        return tuple(self.points)

    def set_points(self,points):
        """ set the points of this polyline """
        self.points = points
        self.invalidate()

    def transform_points( self ):
        """ get the transformed points """
        if not self.transformed_points or self.modified:
            t_points = []
            for x,y in self.points:
                t_points.append(self.transform(x,y))
            self.transformed_points = t_points
        return self.transformed_points

    def get_bbox( self ):
        """ get the bounding box """
        if (not self.transformed_points or self.modified) and self.canvas:
            if not self.bbox:
                self.bbox = Bbox()
            self.bbox.x0,self.bbox.y0,self.bbox.x1,self.bbox.y1 = self.canvas.get_bounds(self.transform_points())
            self.modified = False
        return self.bbox

    def render(self):
        """ override rendering to draw requested polyline """
        self.canvas.polyline(self.transform_points(),self.color)

class Polygon(PolyLine):
    """Display list object for a polygon, filled or not filled"""
    def __init__(self,points,color,fill,parent=None):
        """ points is a list of (x,y) tuples in order to be drawn as a closed polygon in color and if fill is True filled with color """
        PolyLine.__init__(self,points,color,parent)
        self.fill = fill

    def get_fill( self ):
        """ return the fill state of the arc """
        return self.fill

    def set_fill( self, fill):
        """ set the filled state of the arc """
        self.fill = fill
        self.invalidate()

    def render(self):
        """ override rendering to draw requested polygon """
        self.canvas.polygon(self.transform_points(),self.color,self.fill)



# Copyright 2017 James P Goodwin a package to draw graphs of data in a terminal window
""" module that implements graphs of data rendered in a terminal window """
import locale
locale.setlocale(locale.LC_ALL,"")
import curses
import curses.ascii
import sys
import os
import math
from char_draw import canvas
from char_draw import display_list
from data_sources import data_table

def float_range(start, stop, step):
    while start <= stop:
        yield float(start)
        start += step

class GraphSeries():
    """ wrapper for data series to be graphed """
    def __init__(self,data_table,coordinate,color):
        """ constructor takes a data table and a coordinate to choose the column """
        self.data = data_table
        self.column = coordinate
        self.color = color

class Graph(display_list.DisplayList):
    """Graph base class for all graph types """

    def __init__(self,data_table=None,x_values=None,y_values=None,y_unit="",parent=None,canvas=None,top=0,title=None):
        """ base constructor for all graph types constructor takes a data_table which contains values to be graphed,
        x_values is a column reference in the data table for the xaxis values,
        y_values is a list of column references to series of numerical data to be graphed,
        parent is a reference to an enclosing display list,
        canvas is a reference to a canvas to render on """
        display_list.DisplayList.__init__(self,parent,None,canvas)
        self.colors = []
        self.data = data_table
        self.x_values_name = x_values
        self.y_values_names = y_values
        self.y_unit = y_unit
        self.x_values = None
        self.y_values = []
        self.top = top
        self.graph_title = title
        self.initialized = False

    def init(self):
        """ set internal state to default state """
        if not self.initialized:
            if self.canvas:
                self.colors = [self.canvas.cyan, self.canvas.green, self.canvas.red, self.canvas.white]
                self.x_values = GraphSeries(self.data, self.x_values_name, self.canvas.green )
                self.y_values = [GraphSeries(self.data, sy,self.colors[self.y_values_names.index(sy)%len(self.colors)]) for sy in self.y_values_names]
                self.data.listen(self.data_changed)

    def get_series_unit( self ):
        """ get the name of the units for the y axis """
        return self.y_unit

    def get_series( self ):
        """ return list of GraphSeries objects that describe the data series to be graphed """
        return self.y_values

    def get_xvalues( self ):
        """ return the series for the x-values """
        return self.x_values

    def get_data(self):
        """ return reference to the data table """
        return self.data

    def data_changed(self,data_table):
        """ listener that gets called if the data table is changed """
        self.modified = True

    def refresh_data(self):
        """ force a refresh on the data backing this graph """
        self.data.refresh()

    def render(self):
        """ override render to force initialization and layout"""
        if self.canvas:
            self.init()
            self.get_bbox()
            display_list.DisplayList.render(self)

    def get_top_indexes(self):
        """ get the indexes of the self.top items in the series """
        top_values = []
        for series in self.y_values:
            column = series.data.get_column(series.column)
            for idx in range(column.size()):
                top_values.append((column.get(idx).get_float_value(),idx))
        top_values.sort(reverse=True)
        top_indexes = []
        for idx in range(self.top):
            top_indexes.append(top_values[idx][1])
        return top_indexes

    def get_title( self ):
        """ return the title for this graph """
        if self.graph_title:
            return self.graph_title
        else:
            return self.get_data().get_name()

    def is_top(self):
        """ are we justing just the self.top highest items """
        return self.top > 0

class GraphElement( display_list.DisplayList ):
    """ base of all of the graph children """
    def __init__(self, parent, x=0.0, y=0.0, width=0.0, height=0.0):
        display_list.DisplayList.__init__(self,parent)
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def get_location( self ):
        """ return the location of the graph element """
        return (self.x,self.y)

    def get_size( self ):
        """ return the height and width of the graph element """
        return (self.width, self.height)

    def set_location( self, location ):
        """ set the location of the graph element """
        self.x,self.y = location
        self.modified = True

    def set_size( self, size ):
        """ set the size of the graph element """
        self.width,self.height = size
        self.modified = True

    def get_bbox( self ):
        """ return the bounding box of this element if it has no children make it based on it's height and width """
        ret = display_list.DisplayList.get_bbox(self)
        if not ret:
            ret = display_list.Bbox(self.x,self.y,self.x+self.width,self.y+self.height)
        return ret


class GraphTitle(GraphElement):
    """ A title to be displayed on a graph """
    def __init__(self,parent,text=None):
        GraphElement.__init__(self,parent)
        self.text = text
        self.add_child(display_list.Text(0,0,self.text,self.canvas.white))

    def get_bbox(self):
        """ recompute and relayout the component and return it's bbox """
        if self.modified:
            new_children = []
            x,y = self.get_location()
            width,height = self.get_size()
            rows,cols = self.canvas.to_rowcol(width,height)
            r_height,r_width = self.canvas.from_rowcol(1,1)
            text = self.text.strip()
            while text and rows:
                if len(text) > cols:
                    s_idx = cols
                    while s_idx > 0:
                        if text[s_idx] == ' ':
                            break
                        s_idx -= 1
                    if s_idx <= 0:
                        s_idx = cols
                else:
                    s_idx = cols
                new_children.append(display_list.Text(x,y,text[:s_idx],self.canvas.white|(curses.A_STANDOUT if self.is_focus() else curses.A_NORMAL)))
                rows -= 1
                y += r_height
                text = text[s_idx:].lstrip()
            self.set_children(new_children)
        return GraphElement.get_bbox(self)

class GraphLegend(GraphElement):
    """ A legend to be displayed on a graph """
    def __init__(self,parent,listoftext=None):
        GraphElement.__init__(self,parent)
        self.series_labels = listoftext
        r_height,r_width = self.canvas.from_rowcol(1,1)
        x = 0
        y = 0
        for l in self.series_labels:
            self.add_child(display_list.Text(x,y,l,self.canvas.white))
            y += r_height

    def get_bbox(self):
        """ recompute and relayout the component and return it's bbox """
        if self.modified:
            new_children = []
            x,y = self.get_location()
            width,height = self.get_size()
            rows,cols = self.canvas.to_rowcol(width-4,height)
            r_height,r_width = self.canvas.from_rowcol(1,1)
            for text,color in self.series_labels:
                new_children.append(display_list.Rect(x,y,x+1,y+1,color))
                text = text.strip()
                while text and rows:
                    if len(text) > cols:
                        s_idx = cols
                        while s_idx > 0:
                            if text[s_idx] == ' ':
                                break
                            s_idx -= 1
                        if s_idx <= 0:
                            s_idx = cols
                    else:
                        s_idx = cols
                    new_children.append(display_list.Text(x+4,y,text[:s_idx],self.canvas.white))
                    rows -= 1
                    y += r_height
                    text = text[s_idx:].lstrip()
            self.set_children(new_children)
        return GraphElement.get_bbox(self)

class GraphXAxisTitle(GraphTitle):
    """ A title to be displayed on the x-axis """
    def __init(self,parent,text):
        GraphTitle.__init__(self,parent,text)

class GraphYAxisTitle(GraphTitle):
    """ A title to be displayed on the y-axis """
    def __init(self,parent,text):
        GraphTitle.__init__(self,parent,text)

class GraphXAxis(GraphElement):
    """ An X-axis to be displayed on a graph """
    def __init__(self,parent, horizontal = True):
        GraphElement.__init__(self,parent)
        self.horizontal = horizontal
        self.range_min = -1
        self.range_max = -1
        self.sx = 0
        self.dx = 0
        self.values = []
        self.add_child(display_list.PolyLine([],self.canvas.green))

    def get_range( self ):
        """ return range_min, range_max, sx """
        return (self.range_min,self.range_max, self.sx, self.dx)

    def get_values( self ):
        """ return vector of x (value,label) """
        return self.values

    def get_bbox( self ):
        """ compute the bounding box """
        if self.modified:
            new_children = []
            x_values = self.parent.get_xvalues()
            column = x_values.data.get_column(x_values.column)
            type = None

            for c in data_table.ColumnIterator(column):
                if not type or type == data_table.blank_type:
                    type = c.get_type()
                elif type != c.get_type():
                    type = 'mixed'
            self.range_min = -1
            self.range_max = -1
            self.sx = 0
            self.dx = 0
            self.values = []
            if self.parent.is_top():
                top_indexes = self.parent.get_top_indexes()

            if type in [data_table.blank_type,data_table.string_type, 'mixed']:
                if self.parent.is_top():
                    self.range_min = 0
                    self.range_max = len(top_indexes)
                    ii = 0
                    for ix in top_indexes:
                        c = column.get(ix)
                        self.values.append((ii,str(c)))
                        ii += 1
                else:
                    self.range_min = 0
                    self.range_max = column.size()
                    ix = 0
                    for c in data_table.ColumnIterator(column):
                        self.values.append((ix,str(c)))
                        ix += 1
            else:
                self.range_min = None
                self.range_max = None
                if self.parent.is_top():
                    for ix in top_indexes:
                        c = column.get(ix)
                        value = c.get_float_value()
                        if self.range_min == None or value < self.range_min:
                            self.range_min = value
                        if self.range_max == None or value > self.range_max:
                            self.range_max = value
                        self.values.append((value,str(c)))
                else:
                    for c in data_table.ColumnIterator(column):
                        value = c.get_float_value()
                        if self.range_min == None or value < self.range_min:
                            self.range_min = value
                        if self.range_max == None or value > self.range_max:
                            self.range_max = value
                        self.values.append((value,str(c)))

                    if len(self.values) >= 2:
                        extra_value = self.values[-1][0] + (self.values[-1][0] - self.values[-2][0])
                        self.values.append((extra_value," "))
                        if extra_value > self.range_max:
                            self.range_max = extra_value

            r_height,r_width = self.canvas.from_rowcol(1,1)
            width,height = self.get_size()
            ox,oy = self.get_location()
            self.sx = width / max(1.0,(self.range_max-self.range_min))

            x = ox
            y = oy
            new_children.append(display_list.Rect(x,y,x+width,y+height,self.canvas.black,fill=True))
            points = [(x,y)]
            labels = []
            prev_scaled_x = None
            total_dx = 0
            prev_label = None
            for v,label in self.values:
                if label.endswith(".00"):
                    label=label[:-3]
                force = False
                if prev_label:
                    prev_parts = prev_label.split(' ')
                    cur_parts = label.split(' ')
                    new_label = ''
                    no_match = False
                    idx = 0
                    while idx < min(len(prev_parts),len(cur_parts)):
                        if no_match or prev_parts[idx] != cur_parts[idx]:
                            if idx == 0:
                                force = True
                            new_label += cur_parts[idx] + ' '
                            no_match = True
                        idx += 1
                    while idx < len(cur_parts):
                        new_label += cur_parts[idx] + ' '
                        idx += 1
                    prev_label = label
                    label = new_label
                else:
                    prev_label = label

                scaled_x = (v-self.range_min)*self.sx
                if prev_scaled_x != None:
                    total_dx += (scaled_x - prev_scaled_x)
                prev_scaled_x = scaled_x
                t_x,t_y = self.canvas.round_text_position(ox+scaled_x,y+1)
                if t_x >= x or force:
                    points.append((t_x,y))
                    points.append((t_x,t_y))
                    points.append((t_x,y))
                    labels.append((t_x,t_y,label))
                    l_height,l_width = self.canvas.from_rowcol(1,len(label)+1)
                    x = t_x + l_width
            if points[-1][0] < ox+width:
                points.append((ox+width,y))
            self.dx = total_dx / len(self.values)
            new_children.append(display_list.PolyLine(points,self.canvas.green))
            for x,y,label in labels:
                new_children.append(display_list.Text(x,y,label,self.canvas.green))
            self.set_children(new_children)
        return GraphElement.get_bbox(self)

class GraphYAxis(GraphElement):
    """ An X-axis to be displayed on a graph """
    def __init__(self,parent,vertical = True):
        GraphElement.__init__(self,parent)
        self.vertical = vertical
        self.add_child(display_list.PolyLine([],self.canvas.green))
        self.range_min = -1
        self.range_max = -1
        self.sy = 0

    def get_range( self ):
        """ return range_min, range_max, sy """
        return (self.range_min,self.range_max, self.sy)

    def get_bbox(self):
        """ compute the bounding box """
        if self.modified:
            new_children = []
            y_values = self.parent.get_series()
            self.range_min = -1
            self.range_max = -1
            self.dy = 0

            if self.parent.is_top():
                top_indexes = self.parent.get_top_indexes()
                for series in y_values:
                    column = series.data.get_column(series.column)
                    for idx in top_indexes:
                        c = column.get(idx)
                        value = c.get_float_value()
                        if self.range_min < 0 or value < self.range_min:
                            self.range_min = value
                        if self.range_max < 0 or value > self.range_max:
                            self.range_max = value
            else:
                for series in y_values:
                    column = series.data.get_column(series.column)
                    for c in data_table.ColumnIterator(column):
                        value = c.get_float_value()
                        if self.range_min < 0 or value < self.range_min:
                            self.range_min = value
                        if self.range_max < 0 or value > self.range_max:
                            self.range_max = value

            if self.range_min <= 0:
                self.range_min = 0
            if self.range_max <= 0:
                self.range_max = 1
            tick_size = max(1.0,(self.range_max-self.range_min))/5
            if tick_size > 1:
                tick_size = round(tick_size)
            self.range_max += tick_size
            r_height,r_width = self.canvas.from_rowcol(1,1)
            width,height = self.get_size()
            ox,oy = self.get_location()
            new_children.append(display_list.Rect(ox,oy,ox+width,oy+height,self.canvas.black,fill=True))
            oy += height
            ox += (width- 1)
            self.sy = height / max(1.0,(self.range_max-self.range_min))
            x = ox
            y = oy
            points = [(x,y)]
            labels = []
            for data_y in float_range(self.range_min,self.range_max,tick_size):
                scaled_y = (data_y-self.range_min)*self.sy
                label = data_table.format_float(float(data_y))
                if label.endswith(".00"):
                    label=label[:-3]
                t_x,t_y = self.canvas.round_text_position(x-3,oy-scaled_y)
                if t_y <= y:
                    l_height,l_width = self.canvas.from_rowcol(1,len(label)+1)
                    points.append((x,t_y))
                    points.append((t_x,t_y))
                    points.append((x,t_y))
                    labels.append((t_x-l_width,t_y-(l_height/2),label))
                    y = t_y-l_height

            if points[-1][1] > oy-height:
                points.append((x,oy-height))
            new_children.append(display_list.PolyLine(points,self.canvas.green))
            for x,y,label in labels:
                new_children.append(display_list.Text(x,y,label,self.canvas.green))
            self.set_children(new_children)
        return GraphElement.get_bbox(self)

class GraphArea(GraphElement):
    """ The area behind the graph series to be displayed on a graph """
    def __init__(self,parent):
        GraphElement.__init__(self,parent)
        self.add_child(display_list.Rect(0,0,0,0,self.canvas.white))

    def get_bbox(self):
        """ compute the bounding box """
        if self.modified:
            new_children = []
            x,y = self.get_location()
            width,height = self.get_size()
            new_children.append(display_list.Rect(x,y,x+width,y+height,self.canvas.black,fill=True))
            self.set_children(new_children)
        return GraphElement.get_bbox(self)

class GraphBars(GraphElement):
    """ A bar chart plot of a graph series to be displayed on a graph """
    def __init__(self,parent,series,x_axis,y_axis):
        self.series = series
        self.x_axis = x_axis
        self.y_axis = y_axis
        GraphElement.__init__(self,parent)

    def get_bbox(self):
        """ compute the bounding box """
        if self.modified:
            new_children = []
            if self.x_axis.modified:
                self.x_axis.get_bbox()
            if self.y_axis.modified:
                self.y_axis.get_bbox()
            x_min,x_max,x_scale,x_dx = self.x_axis.get_range()
            y_min,y_max,y_scale = self.y_axis.get_range()
            x_values = self.x_axis.get_values()
            x,y = self.get_location()
            width,height = self.get_size()
            y = y+height
            n_series = len(self.parent.get_series())
            i_series = self.parent.get_series().index(self.series)
            series_x_offset = (n_series - (i_series+1)) * (x_dx / n_series)
            x += series_x_offset

            if self.parent.is_top():
                top_indexes = self.parent.get_top_indexes()

            column = self.series.data.get_column(self.series.column)
            idx = 0
            for c in data_table.ColumnIterator(column):
                if not self.parent.is_top() or idx in top_indexes:
                    y_value = c.get_float_value()
                    if self.parent.is_top():
                        x_value = x_values[top_indexes.index(idx)][0]
                    else:
                        x_value = x_values[idx][0]
                    scaled_x = (x_value-x_min)*x_scale
                    scaled_y = (y_value-y_min)*y_scale
                    new_children.append(display_list.Rect(x+scaled_x-(x_dx/4),y,min(x+scaled_x+(x_dx/4),x+width),y-scaled_y,self.series.color))
                idx += 1
            self.set_children(new_children)

        return GraphElement.get_bbox(self)

class GraphLines(GraphElement):
    """ A line chart plot of a graph series to be displayed on a graph """
    def __init__(self,parent,series,x_axis,y_axis,area=False):
        self.series = series
        self.x_axis = x_axis
        self.y_axis = y_axis
        self.area = area
        GraphElement.__init__(self,parent)

    def get_bbox(self):
        """ compute the bounding box """
        if self.modified:
            new_children = []
            if self.x_axis.modified:
                self.x_axis.get_bbox()
            if self.y_axis.modified:
                self.y_axis.get_bbox()
            x_min,x_max,x_scale,x_dx = self.x_axis.get_range()
            y_min,y_max,y_scale = self.y_axis.get_range()
            x_values = self.x_axis.get_values()
            x,y = self.get_location()
            width,height = self.get_size()
            y = y+height

            column = self.series.data.get_column(self.series.column)
            idx = 0
            points = []
            for c in data_table.ColumnIterator(column):
                y_value = c.get_float_value()
                x_value = x_values[idx][0]
                scaled_x = (x_value-x_min)*x_scale
                scaled_y = (y_value-y_min)*y_scale
                points.append((x+scaled_x,y-scaled_y))
                idx += 1
            if self.area:
                p1 = (points[-1][0],y)
                if p1 not in points:
                    points.append(p1)
                p2 = (x,y)
                if p2 not in points:
                    points.append(p2)
                new_children.append(display_list.Polygon(points,self.series.color,True))
            else:
                new_children.append(display_list.PolyLine(points,self.series.color))
            self.set_children(new_children)

        return GraphElement.get_bbox(self)

class BarGraph(Graph):
    """BarGraph that displays a data table as a bar graph"""
    def __init__(self,data_table=None,x_values=None,y_values=None,y_unit="",parent=None,canvas=None,top=0,title=None):
        """ constructor takes a data_table which contains values to be graphed,
        x_values is a column reference in the data table for the xaxis values,
        y_values is a list of column references to series of numerical data to be graphed,
        parent is a reference to an enclosing display list,
        canvas is a reference to a canvas to render on """
        Graph.__init__(self,data_table,x_values,y_values,y_unit,parent,canvas,top,title=title)
        self.title = None
        self.legend = None
        self.x_axis_title = None
        self.y_axis_title = None
        self.x_axis = None
        self.y_axis = None
        self.chart_area = None
        self.chart_series = []
        if self.canvas:
            self.init()

    def init(self):
        """ create the children for all of the graph components """
        if not self.initialized:
            Graph.init(self)
            self.title = GraphTitle(self,self.get_title())
            self.legend = GraphLegend(self,[(s.column,s.color) for s in self.get_series()])
            self.x_axis_title = GraphXAxisTitle(self,self.get_xvalues().column)
            self.y_axis_title = GraphYAxisTitle(self,self.get_series_unit())
            self.x_axis = GraphXAxis(self, horizontal = True )
            self.y_axis = GraphYAxis(self, vertical = True )
            self.chart_area = GraphArea(self)
            self.chart_series = [GraphBars(self,series,self.x_axis,self.y_axis) for series in self.get_series()]
            self.add_child(self.title)
            self.add_child(self.legend)
            self.add_child(self.x_axis_title)
            self.add_child(self.y_axis_title)
            self.add_child(self.chart_area)
            for cs in self.chart_series:
                self.add_child(cs)
            self.add_child(self.x_axis)
            self.add_child(self.y_axis)
            self.initialized = True

    def set_focus( self, state ):
        """ set focus to this graph, in this case tell the title about it so it will highlight """
        if self.title:
            self.title.set_focus(state)
        Graph.set_focus(self,state)

    def get_bbox(self):
        """ arrange the children of the graph based on size of graph """

        if self.modified:
            min_x,min_y,max_x,max_y = self.bounds()

            width = (max_x-min_x) - 4
            height = (max_y-min_y) - 4

            x = 2
            y = 2
            self.title.set_location((x,y))
            self.title.set_size((width,height*0.10))
            y = y + height*0.10
            self.legend.set_location((x,y))
            self.legend.set_size((width,height*0.10))
            y = y + height*0.10
            self.y_axis_title.set_location((x,y+height*0.35))
            self.y_axis_title.set_size((width*0.05,height*0.35))
            self.y_axis.set_location((x+(width*0.05),y))
            self.y_axis.set_size((width*0.05,height*0.70))
            self.x_axis.set_location((x+(width*0.10),y+height*0.70))
            self.x_axis.set_size((width*0.80,height*0.07))
            ya_x,ya_y = self.y_axis.get_location()
            ya_w,ya_h = self.y_axis.get_size()
            xa_x,xa_y = self.x_axis.get_location()
            xa_w,xa_h = self.x_axis.get_size()

            self.chart_area.set_location((ya_x+ya_w,ya_y))
            self.chart_area.set_size((xa_w,ya_h))
            for cs in self.chart_series:
                cs.set_location((ya_x+ya_w+1,ya_y))
                cs.set_size((xa_w,ya_h))

            self.x_axis_title.set_location((xa_x,xa_y+xa_h))
            self.x_axis_title.set_size((xa_w,xa_h))

        return Graph.get_bbox(self)

class LineGraph(Graph):
    """LineGraph that displays a data table as a line graph"""
    def __init__(self,data_table=None,x_values=None,y_values=None,y_unit="",parent=None,canvas=None,area=False,title=None):
        """ constructor takes a data_table which contains values to be graphed,
        x_values is a column reference in the data table for the xaxis values,
        y_values is a list of column references to series of numerical data to be graphed,
        parent is a reference to an enclosing display list,
        canvas is a reference to a canvas to render on """
        Graph.__init__(self,data_table,x_values,y_values,y_unit,parent,canvas,title=title)
        self.title = None
        self.legend = None
        self.x_axis_title = None
        self.y_axis_title = None
        self.x_axis = None
        self.y_axis = None
        self.chart_area = None
        self.chart_series = []
        self.area = area
        if self.canvas:
            self.init()

    def init(self):
        """ create the children for all of the graph components """
        if not self.initialized:
            Graph.init(self)
            self.title = GraphTitle(self,self.get_title())
            self.legend = GraphLegend(self,[(s.column,s.color) for s in self.get_series()])
            self.x_axis_title = GraphXAxisTitle(self,self.get_xvalues().column)
            self.y_axis_title = GraphYAxisTitle(self,self.get_series_unit())
            self.x_axis = GraphXAxis(self, horizontal = True )
            self.y_axis = GraphYAxis(self, vertical = True )
            self.chart_area = GraphArea(self)
            self.chart_series = [GraphLines(self,series,self.x_axis,self.y_axis,area=self.area) for series in self.get_series()]
            self.add_child(self.title)
            self.add_child(self.legend)
            self.add_child(self.x_axis_title)
            self.add_child(self.y_axis_title)
            self.add_child(self.chart_area)
            for cs in self.chart_series:
                self.add_child(cs)
            self.add_child(self.x_axis)
            self.add_child(self.y_axis)
            self.initialized = True

    def set_focus( self, state ):
        """ set focus to this graph, in this case tell the title about it so it will highlight """
        if self.title:
            self.title.set_focus(state)
        Graph.set_focus(self,state)

    def get_bbox(self):
        """ arrange the children of the graph based on size of graph """

        if self.modified:
            min_x,min_y,max_x,max_y = self.bounds()

            width = (max_x-min_x) - 4
            height = (max_y-min_y) - 4

            x = 2
            y = 2
            self.title.set_location((x,y))
            self.title.set_size((width,height*0.10))
            y = y + height*0.10
            self.legend.set_location((x,y))
            self.legend.set_size((width,height*0.10))
            y = y + height*0.10
            self.y_axis_title.set_location((x,y+height*0.35))
            self.y_axis_title.set_size((width*0.05,height*0.35))
            self.y_axis.set_location((x+(width*0.05),y))
            self.y_axis.set_size((width*0.05,height*0.70))
            self.x_axis.set_location((x+(width*0.10),y+(height*0.70)+1.0))
            self.x_axis.set_size((width*0.80,height*0.07))

            ya_bbox = self.y_axis.get_bbox()
            yat_bbox = self.y_axis_title.get_bbox()
            if ya_bbox.x0 < yat_bbox.x1:
                self.y_axis.set_location((self.canvas.round_text_x_position(yat_bbox.x1),ya_bbox.y0))
                self.y_axis.set_size(((ya_bbox.x1-ya_bbox.x0),height*0.70))

            ya_x,ya_y = self.y_axis.get_location()
            ya_w,ya_h = self.y_axis.get_size()

            xa_x,xa_y = self.x_axis.get_location()
            xa_w,xa_h = self.x_axis.get_size()
            self.x_axis.set_location((ya_x+ya_w,ya_y+ya_h))
            self.x_axis.set_size((xa_w - ((ya_x+ya_w)-xa_x),xa_h))

            xa_x,xa_y = self.x_axis.get_location()
            xa_w,xa_h = self.x_axis.get_size()

            self.chart_area.set_location((ya_x+ya_w,ya_y))
            self.chart_area.set_size((xa_w,ya_h))
            for cs in self.chart_series:
                cs.set_location((ya_x+ya_w+1,ya_y))
                cs.set_size((xa_w,ya_h))

            self.x_axis_title.set_location((xa_x,xa_y+xa_h))
            self.x_axis_title.set_size((xa_w,xa_h))

        return Graph.get_bbox(self)


class GraphSlices(GraphElement):
    """ A pie chart plot of a graph series to be displayed on a graph """
    def __init__(self,parent,series):
        self.series = series
        GraphElement.__init__(self,parent)
        self.label_positions = []
        self.angle = 0

    def get_bbox(self):
        """ compute the bounding box """
        if self.modified:
            label_series = self.parent.get_xvalues()
            label_column = label_series.data.get_column(label_series.column)
            data_column = self.series.data.get_column(self.series.column)
            x,y = self.get_location()
            width,height = self.get_size()

            data_idxs = []
            data_total = 0
            for idx in range(0,data_column.size()):
                cell = data_column.get(idx)
                value = cell.get_float_value()
                data_total += value
                data_idxs.append((value,idx,str(label_column.get(idx))))

            data_idxs.sort(reverse=True)

            if data_total > 0:
                total_included = 0
                for idx in range(0,len(data_idxs)):
                    if data_idxs[idx][0]/data_total < 0.02:
                        remaining = len(data_idxs) - idx
                        remaining_idx = data_idxs[idx][1]
                        data_idxs = data_idxs[:idx]
                        data_idxs.append((data_total-total_included,remaining_idx,"%d < 2%%"%(remaining)))
                        break
                    else:
                        total_included += data_idxs[idx][0]


                total_degrees = 0
                self.set_children([])
                label_children = []
                lcx = x
                lcy = y
                l_height,l_width = self.canvas.from_rowcol(1,11)
                px = x+l_width
                pw = width - l_width
                py = y+l_height
                ph = height - (l_height*2)
                for idx in range(0,len(data_idxs)):
                    data_percent = (data_idxs[idx][0]/data_total)
                    data_degrees = round(360.0 * data_percent )
                    slice_color = curses.color_pair(self.canvas.color_min+(30+(idx*11)%((self.canvas.color_max-30)-self.canvas.color_min)))
                    arc = display_list.Arc(px+pw/2,
                                           py+ph/2,
                                           min(ph,pw)/2,
                                           total_degrees,
                                           total_degrees+data_degrees,
                                           slice_color,
                                           True)
                    self.add_child(arc)
                    label = "(%.0f%%) %s"%(data_percent*100.00, data_idxs[idx][2][:11])
                    if lcy < y + height:
                        label_children.append(display_list.Text(lcx,lcy,label,slice_color))
                    lcy += l_height
                    total_degrees += data_degrees

                title = self.series.column
                l_height,l_width = self.canvas.from_rowcol(1,len(title))
                lcx = x+width/2 - l_width/2
                lcy = y+height-l_height
                label_children.append(display_list.Text(lcx,lcy,title,label_series.color))

                for lc in label_children:
                    self.add_child(lc)

        return GraphElement.get_bbox(self)

class PieGraph(Graph):
    """PieGraph that displays a data table as a pie chart"""
    def __init__(self,data_table=None,pie_labels=None,slice_values=None,parent=None,canvas=None,title=None):
        """ constructor takes a data_table which contains values to be graphed,
        pie_labels is a column reference in the data table for the pie slice labels,
        slice_values is a list of column references to series of numerical data to be graphed,
        parent is a reference to an enclosing display list,
        canvas is a reference to a canvas to render on """
        Graph.__init__(self,data_table,pie_labels,slice_values,"",parent,canvas,title=title)
        self.title = None
        self.chart_area = None
        self.chart_series = []
        if self.canvas:
            self.init()

    def init(self):
        """ create the children for all of the graph components """
        if not self.initialized:
            Graph.init(self)
            self.title = GraphTitle(self,self.get_title())
            self.chart_area = GraphArea(self)
            self.chart_series = [GraphSlices(self,series) for series in self.get_series()]
            self.add_child(self.title)
            self.add_child(self.chart_area)
            for cs in self.chart_series:
                self.add_child(cs)
            self.initialized = True

    def set_focus( self, state ):
        """ set focus to this graph, in this case tell the title about it so it will highlight """
        if self.title:
            self.title.set_focus(state)
        Graph.set_focus(self,state)

    def get_bbox(self):
        """ arrange the children of the graph based on size of graph """

        if self.modified:
            min_x,min_y,max_x,max_y = self.bounds()

            width = (max_x-min_x) - 4
            height = (max_y-min_y) - 4

            x = 2
            y = 2
            self.title.set_location((x,y))
            self.title.set_size((width,height*0.10))
            y = y + height*0.15

            self.chart_area.set_location((x,y))
            self.chart_area.set_size((width,height*0.80))
            n_series = len(self.chart_series)
            if n_series > 1:
                split = 1
                while split*split < n_series:
                    split += 1
                sx = width / split
                sy = (height*0.80) / split

                ix = 0
                gx = x
                gy = y
                for cs in self.chart_series:
                    cs.set_location((gx,gy))
                    cs.set_size((sx,sy))
                    ix += 1
                    if ix == split:
                        ix = 0
                        gx = x
                        gy += sy
                    else:
                        gx += sx
            else:
                for cs in self.chart_series:
                    cs.set_location((x,y))
                    cs.set_size((width,height*0.80))

        return Graph.get_bbox(self)

class GraphTable(GraphElement):
    """ Element that lays out a table of text for the series """
    def __init__(self,parent):
        GraphElement.__init__(self,parent)

    def get_bbox(self):
        """ recompute and relayout the component and return it's bbox """
        if self.modified:
            new_children = []
            x,y = self.get_location()
            width,height = self.get_size()
            row_labels = self.parent.get_xvalues()
            columns = self.parent.get_series()

            n_columns = len(columns)+1
            cw = width / n_columns
            ch = height

            rows,cols = self.canvas.to_rowcol(width,height)
            r_height,r_nothing = self.canvas.from_rowcol(1,1)
            col_width = cols // n_columns

            rlc = row_labels.data.get_column(row_labels.column)

            def pad( s, width ):
                if len(s) < width:
                    return s+' '*(width-len(s))
                else:
                    return s

            new_children.append(display_list.Text(x,y,pad(row_labels.column[0:col_width],col_width),self.canvas.white|curses.A_STANDOUT))
            for idx in range(len(columns)):
                new_children.append(display_list.Text(x+cw+idx*cw,y,pad(columns[idx].column[0:col_width],col_width),self.canvas.white|curses.A_STANDOUT))
            for ridx in range(min(rows-1,rlc.size())):
                y += r_height
                new_children.append(display_list.Text(x,y,pad(str(rlc.get(ridx))[0:col_width],col_width),self.canvas.white))
                for cidx in range(len(columns)):
                    column = columns[cidx].data.get_column(columns[cidx].column)
                    new_children.append(display_list.Text(x+cw+cidx*cw,y,pad(str(column.get(ridx))[0:col_width],col_width),self.canvas.white))

            self.set_children(new_children)

            return GraphElement.get_bbox(self)

class TableGraph(Graph):
    """TableGraph that displays a data table as a formatted table"""
    def __init__(self,data_table=None,row_labels=None,column_values=None,parent=None,canvas=None,title=None):
        """ constructor takes a data_table which contains values to be graphed,
        row_labels is a column reference in the data table for the table row labels,
        column_values is a list of column references to series of numerical or text data to be displayed in the table,
        parent is a reference to an enclosing display list,
        canvas is a reference to a canvas to render on """
        Graph.__init__(self,data_table,row_labels,column_values,"",parent,canvas,title=title)
        self.title = None
        self.chart_area = None
        self.graph_table = None
        if self.canvas:
            self.init()

    def init(self):
        """ create the children for all of the graph components """
        if not self.initialized:
            Graph.init(self)
            self.title = GraphTitle(self,self.get_title())
            self.chart_area = GraphArea(self)
            self.graph_table = GraphTable(self)
            self.add_child(self.title)
            self.add_child(self.chart_area)
            self.add_child(self.graph_table)
            self.initialized = True

    def set_focus( self, state ):
        """ set focus to this graph, in this case tell the title about it so it will highlight """
        if self.title:
            self.title.set_focus(state)
        Graph.set_focus(self,state)

    def get_bbox(self):
        """ arrange the children of the graph based on size of graph """

        if self.modified:
            min_x,min_y,max_x,max_y = self.bounds()

            width = (max_x-min_x) - 4
            height = (max_y-min_y) - 4

            x = 2
            y = 2
            self.title.set_location((x,y))
            self.title.set_size((width,height*0.10))
            y = y + height*0.15

            self.chart_area.set_location((x,y))
            self.chart_area.set_size((width,height*0.80))

            self.graph_table.set_location((x,y))
            self.graph_table.set_size((width,height*0.80))

        return Graph.get_bbox(self)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("VERSION", "r") as fh:
    version = fh.read().strip()

setup(
    name='terminal-dashboard',
    version=version,
    url='http://github.com/jpfxgood/dashboard',
    author="James Goodwin",
    author_email="dashboard@jlgoodwin.com",
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    description='A tool for showing a customizable graphical dashboard of various data sources in a terminal',
    long_description_content_type='text/markdown',
    long_description=long_description,
    license = 'MIT',
    keywords= [
        'dashboard',
        'monitor',
        'terminal',
        'graphs',
    ],
    install_requires=[
        'psutil',
        'elasticsearch',
        'paramiko',
        'keyring',
    ],
    scripts=[
        'scripts/dashboard',
    ],
    packages=[
        'char_draw',
        'dashboard',
        'data_sources',
    ],
    python_requires='>=3.6',
)
